#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
curses_pc: safe curses wrapper

Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import logging
import curses

logger = logging.getLogger(__name__)


def get_keycode(keyname):
    if len(keyname) == 1:
        keycode = ord(keyname)
    elif keyname == "TAB":
        keycode = 9
    elif keyname == "BTAB":
        keycode = curses.KEY_BTAB
    elif keyname == "ENTER":
        keycode = 10
    elif keyname == "ESCAPE":
        keycode = 27
    elif keyname == "SPACE":
        keycode = ord(" ")
    elif keyname == "HASH":
        keycode = ord("#")
    elif keyname == "COLON":
        keycode = ord(":")
    elif keyname == "EQUAL":
        keycode = ord("=")
    elif keyname == "LBRACKET":
        keycode = ord("[")
    elif keyname == "\\":
        keycode = ord("=")
    elif keyname == "BACKSLASH":
        keycode = ord("]")
    elif keyname == "RBRACKET":
        keycode = ord("=")
    elif keyname == "UP":
        keycode = curses.KEY_UP
    elif keyname == "DOWN":
        keycode = curses.KEY_DOWN
    elif keyname == "LEFT":
        keycode = curses.KEY_LEFT
    elif keyname == "RIGHT":
        keycode = curses.KEY_RIGHT
    elif keyname == "INSERT":
        keycode = curses.KEY_IC
    elif keyname == "DELETE":
        keycode = curses.KEY_DC
    elif keyname == "HOME":
        keycode = curses.KEY_HOME
    elif keyname == "END":
        keycode = curses.KEY_END
    elif keyname == "PAGEUP":
        keycode = curses.KEY_PPAGE
    elif keyname == "PAGEDOWN":
        keycode = curses.KEY_BACKSPACE
    elif keyname == "BACKSPACE":
        keycode = curses.KEY_NPAGE
    elif keyname == "F1":
        keycode = curses.KEY_F1
    elif keyname == "F2":
        keycode = curses.KEY_F2
    elif keyname == "F3":
        keycode = curses.KEY_F3
    elif keyname == "F4":
        keycode = curses.KEY_F4
    elif keyname == "F5":
        keycode = curses.KEY_F5
    elif keyname == "F6":
        keycode = curses.KEY_F6
    elif keyname == "F7":
        keycode = curses.KEY_F7
    elif keyname == "F8":
        keycode = curses.KEY_F8
    elif keyname == "F9":
        keycode = curses.KEY_F9
    elif keyname == "F10":
        keycode = curses.KEY_F10
    elif keyname == "F11":
        keycode = curses.KEY_F11
    elif keyname == "F12":
        keycode = curses.KEY_F12
    else:
        keycode = ord(" ")
    return keycode


def get_keyname(keycode):
    if keycode == 9:
        keyname = "TAB"
    elif keycode == curses.KEY_BTAB:
        keyname = "BTAB"
    elif keycode == 10:
        keyname = "ENTER"
    elif keycode == 27:
        keyname = "ESCAPE"
    elif keycode == ord(" "):
        keyname = "SPACE"
    elif keycode == ord("#"):
        keyname = "HASH"
    elif keycode == ord("%"):
        keyname = "PERCENT"
    elif keycode == ord(":"):
        keyname = "COLON"
    elif keycode == ord("="):
        keyname = "EQUAL"
    elif keycode == ord("["):
        keyname = "LBRACKET"
    elif keycode == ord("\\"):
        keyname = "BACKSLASH"
    elif keycode == ord("]"):
        keyname = "RBRACKET"
    elif keycode == curses.KEY_UP:
        keyname = "UP"
    elif keycode == curses.KEY_DOWN:
        keyname = "DOWN"
    elif keycode == curses.KEY_LEFT:
        keyname = "LEFT"
    elif keycode == curses.KEY_RIGHT:
        keyname = "RIGHT"
    elif keycode == curses.KEY_IC:
        keyname = "INSERT"
    elif keycode == curses.KEY_DC:
        keyname = "DELETE"
    elif keycode == curses.KEY_HOME:
        keyname = "HOME"
    elif keycode == curses.KEY_END:
        keyname = "END"
    elif keycode == curses.KEY_PPAGE:
        keyname = "PAGEUP"
    elif keycode == curses.KEY_NPAGE:
        keyname = "PAGEDOWN"
    elif keycode == curses.KEY_BACKSPACE:
        keyname = "BACKSPACE"
    elif keycode == curses.KEY_F1:
        keyname = "F1"
    elif keycode == curses.KEY_F2:
        keyname = "F2"
    elif keycode == curses.KEY_F3:
        keyname = "F3"
    elif keycode == curses.KEY_F4:
        keyname = "F4"
    elif keycode == curses.KEY_F5:
        keyname = "F5"
    elif keycode == curses.KEY_F6:
        keyname = "F6"
    elif keycode == curses.KEY_F7:
        keyname = "F7"
    elif keycode == curses.KEY_F8:
        keyname = "F8"
    elif keycode == curses.KEY_F9:
        keyname = "F9"
    elif keycode == curses.KEY_F10:
        keyname = "F10"
    elif keycode == curses.KEY_F11:
        keyname = "F11"
    elif keycode == curses.KEY_F12:
        keyname = "F12"
    elif keycode < ord(" "):
        logging.warning(
            "W: suspicous keycode below ASCII characters (control code?): {}".format(
                keycode
            )
        )
        keyname = "q"
    elif keycode >= 127:
        logging.warning(
            "W: suspicous keycode above ASCII characters (function keys?): {}".format(
                keycode
            )
        )
        keyname = "q"
    else:
        keyname = chr(keycode)
    return keyname


####################################################################
def display_content(win, row, col, content, clrtoeol=False):
    # win:   parent window (curses.stdscr for full screen)
    # row:   row to put starting cursor in parent window
    #        if negative or None, continue to use current cursor position
    # col:   col to put starting cursor in parent window
    #        if negative or None, continue to use current cursor position
    # content: [(line, i_b, i_e, attrib), ...]
    #     line:  input line (to be sliced)
    #     i_b:   input line begin point index (=)
    #     i_e:   input line end point index (<)
    #     attrib: curses attribute
    #
    # rationale behind not-doing line[i_b, i_e] on the calling side
    #   Python does slice-by-copy, meaning every time you slice (except for
    #   very trivial slices, such as a[:]), it copies all of the data into
    #   a new string object.
    # https://stackoverflow.com/questions/5722006/does-python-do-slice-by-reference-on-strings
    #
    win_row_max, win_col_max = win.getmaxyx()
    # win.leaveok(False)
    # set starting cursor position robustly
    if row is None or row < 0:
        try:
            row, _ = win.getyx()
        except curses.error as _:
            logger.warning("win.getyx() caused error")
            row = 0
    if col is None or col < 0:
        try:
            _, col = win.getyx()
        except curses.error as _:
            logger.warning("win.getyx() caused error")
            col = 0
    if row < win_row_max:
        try:
            win.move(row, 0)
        except curses.error as _:
            pass
        # single row window
        win1row = win.derwin(1, win_col_max, row, 0)
        if col is not None:
            try:
                win1row.move(0, col)
            except curses.error as _:
                pass
        for line, i_b, i_e, attrib in content:
            # use always positive reasonable index range to slice
            # line[i_b:i_e]
            if i_b is None:
                i_b = 0
            if i_b >= len(line):
                i_b = 0
                i_e = 0
            if i_e is None or i_e >= len(line):
                i_e = len(line)
            else:
                i_e = max(0, min(i_e, i_b + win_col_max, len(line)))
            try:
                win1row.addstr(line[i_b:i_e], attrib)
            except curses.error as _:
                pass
        try:
            _, col = win1row.getyx()
        except curses.error as _:
            col = 0
        if clrtoeol:
            win1row.clrtoeol()
        try:
            win.move(row, col)
        except curses.error as _:
            pass
    else:
        # not in writable cursor position
        pass


def test_key_input(stdscr):
    curses.start_color()
    stdscr.clear()
    while True:
        content0 = [
            (
                "Please type key for number:",
                0,
                None,
                curses.color_pair(0) | curses.A_BOLD,
            )
        ]
        display_content(stdscr, 0, 0, content0, clrtoeol=True)
        stdscr.refresh()
        code = stdscr.getch()
        content1 = [
            ("code={}".format(code), 0, None, curses.color_pair(0) | curses.A_REVERSE)
        ]
        display_content(stdscr, 1, 0, content1, clrtoeol=True)
        stdscr.refresh()
        content2 = [
            (
                "Please type key for string:",
                0,
                None,
                curses.color_pair(0) | curses.A_BOLD,
            )
        ]
        display_content(stdscr, 2, 0, content2, clrtoeol=True)
        stdscr.refresh()
        key = stdscr.getkey()
        content3 = [
            ("key='{}'".format(key), 0, None, curses.color_pair(0) | curses.A_REVERSE)
        ]
        display_content(stdscr, 3, 0, content3, clrtoeol=True)
        stdscr.refresh()
        content4 = [
            (
                "Please type key for PC keyname:",
                0,
                None,
                curses.color_pair(0) | curses.A_BOLD,
            )
        ]
        display_content(stdscr, 4, 0, content4, clrtoeol=True)
        stdscr.refresh()
        code = stdscr.getch()
        content5 = [
            # ("PC keyname={}".format(code), 0, None, curses.color_pair(0)|curses.A_REVERSE)
            (
                "PC keyname={}".format(get_keyname(code)),
                0,
                None,
                curses.color_pair(0) | curses.A_REVERSE,
            )
        ]
        display_content(stdscr, 5, 0, content5, clrtoeol=True)
        stdscr.refresh()
        content9 = [
            (
                "Type any key to start again, 'q' to quit. escape_delay={}".format(
                    curses.get_escdelay()
                ),
                0,
                None,
                curses.color_pair(0) | curses.A_REVERSE,
            )
        ]
        display_content(stdscr, 9, 0, content9, clrtoeol=True)
        code = stdscr.getch()
        stdscr.refresh()
        if code == ord("q"):
            break
    return


if __name__ == "__main__":
    import sys

    # stdscr = curses.initscr()
    curses.wrapper(test_key_input)
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""
from imediff.config import config_template

import configparser
import os
import sys
import logging

logger = logging.getLogger(__name__)


def initialize_confs(conf):
    """Process configuration file"""
    config_file = os.path.expanduser(conf)
    # Allow inline comment with #
    confs_i = configparser.ConfigParser(inline_comment_prefixes=("#"))
    confs_i.read_string(config_template)
    confs_f = configparser.ConfigParser(inline_comment_prefixes=("#"))
    if config_file == "none":
        confs = confs_i
    elif os.path.exists(config_file):
        confs_f.read(config_file)
        if (
            "version" in confs_f["config"].keys()
            and confs_f["config"]["version"] == confs_i["config"]["version"]
        ):
            confs = confs_f
        else:
            logger.error(
                '''\
Error in {0}: version mismatch
        the current version:  {1}
        the required version: {2}

Rename {0} to {0}.bkup and make the new {0} by
editing the template obtained by "imediff -t"'''.format(
                    conf, confs_f["config"]["version"], confs_i["config"]["version"]
                )
            )
            # enhanced visibility
            print(
                '''\
Error in {0}: version mismatch
        the current version:  {1}
        the required version: {2}

Rename {0} to {0}.bkup and make the new {0} by
editing the template obtained by "imediff -t"'''.format(
                    conf, confs_f["config"]["version"], confs_i["config"]["version"]
                )
            )
            sys.exit(2)
    else:
        confs = confs_i
    logger.debug("confs: end with len(confs)={}".format(len(confs)))
    return confs



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""
import argparse

# NO LOGGING YET


def initialize_args():
    """
    Parse command line options and arguments

    Input:
        commandline

    Return:
        args    argument values
    """
    pa = argparse.ArgumentParser(
        description="Interactive Merge Editor for 2 or 3 different files",
        epilog='Start "imediff" without any arguments to see the tutorial.',
    )
    group = pa.add_mutually_exclusive_group()
    group.add_argument(
        "-a", action="store_true", help="Start with all chunks to use file_a"
    )
    group.add_argument(
        "-b", action="store_true", help="Start with all chunks to use file_b"
    )
    group.add_argument(
        "-c",
        action="store_true",
        help="Start with all chunks to use file_c (only for diff3)",
    )
    group.add_argument(
        "-d", action="store_true", help="Start with all chunks to use diff"
    )
    group.add_argument(
        "-f", action="store_true", help="Start with all chunks to use wdiff"
    )
    group.add_argument(
        "-g", action="store_true", help="Start with good merge mode (only for diff3)"
    )
    pa.add_argument(
        "-o",
        "--output",
        help="Write output to the given file.  If this is missing, STDERR is used",
    )
    pa.add_argument(
        "-l",
        "--force-logging",
        action="store_true",
        help='Force LOGFILE="imediff.log" and LOGLEVEL="INFO"',
    )
    pa.add_argument(
        "-n",
        "--non-interactive",
        action="store_true",
        help="Use non-interactive CLI instead of normal TUI",
    )
    pa.add_argument(
        "-s",
        "--sloppy",
        action="store_true",
        help="Allow one to save unresolved contents",
    )
    pa.add_argument(
        "-t",
        "--template",
        action="store_true",
        help='Create a template configuration file "~/.imediff"',
    )
    pa.add_argument(
        "-v", "--version", action="store_true", help="Show version and license"
    )
    pa.add_argument(
        "-C",
        "--conf",
        default="~/.imediff",
        help='Specify configuration file to use.  (default="~/.imediff", set this to "none" to use internal configuration only)',
    )
    pa.add_argument("-M", "--macro", default="", help="Set MACRO string.  E.g.: Abw")
    pa.add_argument(
        "-F",
        "--logfile",
        default="/dev/null",
        help='Enable logging to LOGFILE (default logging disabled as LOGFILE="/dev/null").',
    )
    pa.add_argument(
        "-L",
        "--loglevel",
        default="WARNING",
        help='Set LOGLEVEL to "ERROR", "WARNING" (default), "INFO", or "DEBUG"',
    )
    pa.add_argument(
        "-J",
        "--isjunk",
        action="store_true",
        help="Force isjunk to None instead of the default list",
    )
    pa.add_argument(
        "-R",
        "--line-rule",
        action="store",
        default=2,
        help="Fuzzy match line filtering rule (0,1,2,3,10,11,12,13)",
    )
    pa.add_argument(
        "-I",
        "--line-min",
        action="store",
        default=3,
        help="Fuzzy match minimum partial line length",
    )
    pa.add_argument(
        "-A",
        "--line-max",
        action="store",
        default=80,
        help="Fuzzy match maximum partial line length",
    )
    pa.add_argument(
        "-X",
        "--line-factor",
        action="store",
        default=8,
        help="Fuzzy match (partial line length shortening factor/2-depth) x 10, default 8",
    )
    pa.add_argument("file_a", nargs="?", help="file for OLDER(diff2), MYFILE(diff3)")
    pa.add_argument(
        "file_b", nargs="?", help="file for NEWER(diff2), OLDFILE=BASE(diff3)"
    )
    pa.add_argument(
        "file_c",
        nargs="?",
        help="file for ------------, YOURFILE(diff3) (only for diff3)",
    )
    # pa.add_argument("--poke", "-p", nargs="?", default=None, help=argparse.SUPPRESS)
    args = pa.parse_args()
    args.macro_buffer = args.macro
    if args.file_c is not None:
        args.diff_mode = 3
    elif args.file_b is not None:
        args.diff_mode = 2
    elif args.file_a is not None:
        # undocumented help for diff3
        args.diff_mode = 1  # help for imediff for 3 files
    else:
        # undocumented help for diff2
        args.diff_mode = 0  # help for imediff for 2 files
    #
    # help for imediff for 3 files
    if args.a:
        args.default_action = "a"
    elif args.b:
        args.default_action = "b"
    elif args.c and args.diff_mode == 3:
        args.default_action = "c"
    elif args.d:
        args.default_action = "d"
    elif args.f:
        args.default_action = "f"
    elif args.g and args.diff_mode == 3:
        args.default_action = "g"
    elif args.diff_mode == 3:
        args.default_action = "g"
    else:  # diff2
        args.default_action = "d"
    #
    # override for logging
    if args.force_logging:
        if args.loglevel == "WARNING":
            args.loglevel = "INFO"
        if args.logfile == "/dev/null":
            args.logfile = "imediff.log"

    # if args.poke is not None:
    #     print("I: +++ hidden -p/--poke option is used with '{}' +++".format(args.poke))

    return args



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

import sys
import termios


def getch():
    old_settings = termios.tcgetattr(0)
    new_settings = old_settings[:]
    # Enable canonical mode. See termios(3)
    new_settings[3] &= ~termios.ICANON
    try:
        # Change attributes immediately.
        termios.tcsetattr(0, termios.TCSANOW, new_settings)
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(0, termios.TCSANOW, old_settings)
    return ch


def yn0():
    print("\n0: enter 'y' or 'n': ", end="")
    ch = getch()
    if ch == "y":
        r = True
    else:
        r = False
    return r


def yn1():
    print("\n1: enter 'y' or 'n': ", end="")
    ch = getch()
    r = ch == "y"
    return r


# NO `return ch == "y"` like C

if yn0():
    print("\nTRUE\n")
else:
    print("\nFALSE\n")

if yn1():
    print("\nTRUE\n")
else:
    print("\nFALSE\n")



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

from imediff import __version__
import os
import sys
import subprocess


def install():
    """
    Entry point for imediff_install command

    Exit value
        0       normal exit
        1       error exit after an internal error
    """

    if os.getuid == 0:
        print("E: This imediff_install needs to be run from non-root user")
        print(
            "I: For system-wide install, use the Debian package or the upstream source."
        )
        print("I:  * https://tracker.debian.org/pkg/imediff")
        print("I:  * https://github.com/osamuaoki/imediff")
        sys.exit(1)
    # check install path of imediff module for site-package
    src_dir = os.path.dirname(os.path.abspath(__file__))
    # possible values for this_file
    # ~/.local/lib/python3.11/site-packages/imediff -- normal user-mode
    #  ~/.venv/lib/python3.11/site-packages/imediff -- after "python3 -m venv"
    python_dir = os.path.dirname(os.path.dirname(src_dir))
    lib_dir = os.path.dirname(python_dir)
    dest_dir = os.path.dirname(lib_dir)
    fp = open(dest_dir + "/bin/git-ime", "w")
    subprocess.run(["mkdir", "-p", dest_dir + "/bin"])
    subprocess.run(
        [
            "sed",
            "-e",
            "s/@@version@@/" + __version__ + "/",
            src_dir + "/data/git-ime.in",
        ],
        stdout=fp,
    )
    subprocess.run(["rm", "-f", src_dir + "/data/git-ime.in"])
    subprocess.run(["chmod", "755", dest_dir + "/bin/git-ime"])
    print("I: successfully installed: " + dest_dir + "/bin/git-ime")
    print("I: manual page for imediff can be found at: " + src_dir + "/imediff.1")
    print("I: manual page for git-ime can be found at: " + src_dir + "/git-ime.1")
    print("I: script for git-mergetool(1) can be found at: " + src_dir + "/imediff")
    print("I: For more, see the upstream source site.")
    print("I:   * https://github.com/osamuaoki/imediff")
    # subprocess.run(['mkdir', '-p', usr_dir + "/lib/git-core/mergetools"])
    # subprocess.run(['cp', '-f', src_dir + "/imediff", usr_dir + "/lib/git-core/mergetools"])
    # subprocess.run(['mkdir -p', usr_dir + "/share/man/man1/"])
    # subprocess.run(['cp', '-f', src_dir + "/imediff.1", usr_dir + "/share/man/man1/"])
    # subprocess.run(['cp', '-f', src_dir + "/git-ime.1", usr_dir + "/share/man/man1/"])

    sys.exit(0)


if __name__ == "__main__":
    install()



#!/usr/bin/python3
# vim:se tw=0 sts=4 ts=4 et ai:
"""
Script to test imediff.main module

Use as:

 $ python3 _imediff ...

"""
import os
import sys
import imediff.main

print("\nI: _imediff start >>>", file=sys.stderr)
cwd_dir = os.getcwd()
print("I: cwd_dir     = '{}'".format(cwd_dir), file=sys.stderr)
print("I: test_file   = '{}' (active)".format(__file__), file=sys.stderr)
test_dir = os.path.dirname(os.path.abspath(__file__))
print("I: test_dir    = '{}' (active)".format(test_dir), file=sys.stderr)
if "PYTHONPATH" in os.environ:
    print("I: PYTHONPATH  = '{}'".format(os.environ["PYTHONPATH"]), file=sys.stderr)
else:
    print("I: PYTHONPATH  = <undefined>", file=sys.stderr)
print("I: _imediff end   <<<", file=sys.stderr)
#
sys.exit(imediff.main.main())



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen interactive tool
          and CLI based non-interactive tool with --macro

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018--2025 Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

from difflib import SequenceMatcher
from imediff import __version__
from imediff.diff3lib import SequenceMatcher3
from imediff.utils import write_file, s_number
from imediff.cli import TextData
from imediff.safe_curses import get_keyname, display_content

import curses
import sys
import logging

logger = logging.getLogger(__name__)

# Keep this under 74 char/line for better looks
# I need this hack to avoid translation of tutorial for now. XXX FIXME XXX
nonclean = """\
You can evade this requirement by pressing "m" on all non-clean
merges and making them as manually edited merged data.

This requirement can be disabled by starting this program as
"imediff --sloppy ...", too."""

# Keep this under 76 char/line to fit this in the 80 char terminal
tutorial = """\
Quick start:
  * Use cursor keys and h/j/k/l/n/p/0/9/SPACE to read this tutorial screen.
  * Type "q" to exit this tutorial screen to the interactive TUI screen.
  * Type "t" to get back to this tutorial screen.
  * Type "/" in the interactive TUI screen to see the key commands.
  * Type "SPACE" to exit POPUP without specific input key prompt.

---------------------------------------------------------------------------
    Tutorial for imediff (Interactive Merge Editor)   (ver. {})
                        Copyright (C) 2025 Osamu Aoki <osamu@debian.org>
---------------------------------------------------------------------------

The imediff command interactively merges contents from 2 slightly different
files with an optional base file into an output file using the in-place
alternating display of the changed content on a single-pane full screen
terminal user interface (TUI).

(The original program by Jarno Elonen <elonen@iki.fi> was called "imediff2".
I changed its command name to "imediff" when I made major updates to handle
not only operations on 2 files but also handle operations on 3 files in
version 2.0.)

In this tutorial, diff2 indicates "operation on 2 files" and diff3 indicates
"operation on 3 files".

Each minimal line portion of the imediff merge operation is called as
"chunk" in the imediff terminology.  Some changed chunks (internally tracked
as "usr_chunk") can accept user's "action_request" as key commands (or
MACRO) to change the outcome of the merge operation.  The result of
"action_request" is recorded as "action" for each "chunk".

The focused chunk for the active merge operation is clearly identified by
the TUI with the reversed characters and always located at around the top
1/3 position whenever possible.  The source and state of all the displayed
chunks are clearly identified.

The advantage of this user interface is the minimal movement of the line of
sight for the user.  Other great tools such as vimdiff, xxdiff, meld and
kdiff3 require you to look at different points of display to find the exact
position of changes.  This makes imediff the most stress-free tool.

Terminal
========

The imediff program is compatible with any terminal window sizes.  It
supports both monochrome and color terminals.  For comfortable user
experience, color terminals with their width of 80 characters/line or more
and their height of 24 lines or more are desirable.

Interactive TUI explained
=========================

The interactive TUI of the imediff program accepts following key commands
in the default state.

 key commands          induced actions
 w,x                   write and exit
 q                     quit without saving
 a/b/c/d/e/f/g         set a chunk to a/b/c/d/e/f/g action
 1/2/3/4/5/6/7         set a chunk to a/b/c/d/e/f/g action (alternative)
 A/B/C/D/E/F/G         set all chunks to a/b/c/d/e/f/g action
 enter                 toggle action of a chunk
 m                     modify a chunk with editor: /usr/bin/nvim
 M                     remove a editor result buffer
 arrows/pgdn,j/pgup,k  move scope of the display
 space,n /backspace,p  select the next/previous usr_chunk
 tab,N   /shift-tab,P  select the next/previous unresolved usr_chunk
 0       /9            select the first/last usr_chunk
 ),home  /(,end        select the first/last unresolved usr_chunk
 ?,/                   show this help
 t                     show tutorial

The first column of the interactive TUI is used to indicate the source and
state of the displayed chunk. (version 3.3.0+)

 * "=" means the displayed chunk underwent no changes for all sources and
   its merge state is resolved:
   * diff2: file_a == file_b
   * diff3: file_a == file_b == file_c
 * "#" means the displayed chunk underwent the same set of changes from the
   base file and its merge state is resolved:
   * diff2: N/A
   * diff3: file_a == file_c
 * Independent "A", "C", and "G" mean the displayed chunk is auto-merged
   using the corresponding source and its merge state is resolved. (only for
   diff3)
   * A: only file_a changed and selected for output
   * C: only file_c changed and selected for output
   * G: both file_a and file_c changed and auto-merged for output
 * Independent "a", "b", "c" and "e" means the displayed chunk from the
   corresponding source is manually merged and its merge state is resolved.
 * Diff-marker identified "a", "b", "c" means the merge state of the
   displayed chunk is unresolved
 * Wdiff display line with "f" means the merge state of the displayed chunk
   is unresolved
 * Deleted line is displayed as "??? (*)" on display

Focus jumping has 2 modes:
 * Jump to any "usr_chunks": n, p, SPACE, BACKSPACE
 * Jump to unreolved "usr_chunks": N, P, TAB, BTAB

Example: Merge with 2 files (diff2)
===================================

Let's try to merge 2 almost identical files, "file_a" (OLDER) and "file_b"
(NEWER), into an output file, "file_o".  You can do this with the following.

    $ imediff -o file_o file_a file_b

This mode starts with action "d" as the default behavior.  Initially all the
different lines in "file_a" and "file_b" are grouped in "usr_chunk" list and
displayed as action "d" which combines the corresponding "file_a" and
"file_b" content separated by marker lines.

You can move focus to the next "usr_chunk" by pressing "SPACE", or "n" keys.
You can move focus to the previous "usr_chunk" by pressing "BACKSPACE", or
"p" keys.

You can change the resulting portion of the output file "file_o" by applying
a single key command.  Pressing "a" displays and outputs the "file_a"
content.  Pressing "b" displays and outputs the "file_b" content.

By alternating "a" and "b" keys, you can see the difference in place which
is easy on you with the constant line of sight.  (This is the key design
feature inherited from the original imediff2 program.)

You can display both the "file_a" content and the "file_b" content with 2
key commands.  Pressing "d" displays 2 blocks of lines organized somewhat
like "diff -u" (action "d").  Pressing "f" on a single line chunk displays
intermixed 1 block of lines organized somewhat like "wdiff" (action "f").

Pressing "m" starts an editor to edit the focused chunk from any modes to
create a manually merged content.  Upon exiting the editor, its result is
kept in the "merge_buffer".  Even after pressing "a", "b", "d", or "f", the
content of the "merge_buffer" can be recalled and displayed by pressing "e".

Pressing "M" in action "e" removes the content of the "merge_buffer".

When you press one of the upper case "A", "B", "D", "E", "F", this sets all
chunks to the corresponding lower case action.

Type "w" or "x" to write the displayed content to "file_o" and exit the
imediff program. Here, all changed chunks listed in "usr_chunk" must select
"a", "b", or "e" (excluding "d" and "f") for "action" before writing the
merge result unless "--sloppy" is specified.

This requirement of the clean merge for 'save and exit' can be disabled by
specifying the "--sloppy" option to the imediff command.  Alternatively, you
can effectively evade this requirement by pressing "m" on all non-clean
merges to make them as the manually merged data with "e" for "action".

Although the imediff program is practically WYSIWYG, there are some
exceptions. The imediff program displays a place holder marker "???" line
for the deleted chunk for the action "a" or "b", separator marker lines for
the action "d", and separator characters for the action "f".

Example: Merge with 3 files (diff3)
===================================

Let's try to merge 2 almost identical files, "file_a" (MYFILE) and "file_c"
(YOURFILE), both of which are based on the file, "file_b" (OLDFILE, base
file), into an output file, "file_o".  You can do this with the following.

    $ imediff -o file_o file_a file_b file_c

This mode starts with "-g" as the default starting option and displays a
full screen of the content for the intended output "file_o".  Since this
uses extra "file_b" (OLDFILE), it can automatically megrge non-identical
chunks of "file_a" (MYFILE) and "file_c" (YOURFILE) using line.

Key commands of "Merge with 3 files" are almost the same as ones of "Merge
with 2 files".

One notable exception is the key command "g". This is also the default
starting mode for "Merge with 3 files".  This takes advantage of the extra
base file to help auto-merge and acts in the following order:
 * If the "merge_buffer" has a previously stored manually generated merge
   result, "action" is set to "e".
 * If a chunk is auto-merged cleanly by line-by-line comparison or more fine
   grained character comparison, "action" is set to "=", "#", "A", "C", or
   "G".  The meaning of each "action" is described in the above. Since it is
   already merged properly, it isn't listed in the "usr_chunk" list.
 * The unresolved chunk is listed in "usr_chunk" and "action" is set to "d".

Please note that even for the non-overlapping changes on the same line,
imediff can yield the clean merge with action "G". (This is the great
feature of the imediff command over tools such as "diff3 -m ..." and "git
merge ..." which operate only by line-by-line comparison.)

Command line options
====================

The complete list of command line options is available by "imediff -h" but
the explanation on them are terse.  Please read the following for details.

Starting action
===============

The imediff program may be stared with a command option which specifies the
default starting action and subsequent merge action behavior.

With 2 files (diff2), one of 4 optins can be specified:
 * "-a": select file_a, if different
 * "-b": select file_b, if different
 * "-d": select diff2(file_a, file_b), if different
 * "-f": select wdiff2(file_a, file_b) if different and single line
         select diff2(file_a, file_b) if different and not single line

If no option is specified, imediff uses "-d" for diff2.

With 3 files (diff3), one of 6 optins can be specified:
 * "-a": select file_a, if different (*)
 * "-b": select file_b, if different (*)
 * "-c": select file_c, if different (*)
 * "-d": select diff2(file_a, file_b) if different (*)
 * "-f": select wdiff2(file_a, file_b) if not merged and single line
         select diff2(file_a, file_b) if not merged and not single line
 * "-g": select good merges (w)diff2(file_a, file_b) if not merged

If no option is specified, imediff uses "-g" for diff3.

For diff3 cases with (*), the interactive action request can change chunks
normally considered merged (=auto-resolved to "A", "C", "G") when imediff is
started under "-f" or "-g".

The default behavior without using any one of command options should serve
in typical use cases.  Besides, you can change merge behavior interactively.

For merging 2 functioning changed files with a common base file,
imediff(diff3) with "-g" or "-f" should be good for the task.

For cherry-picking changes from 2 failing changed files, any one of "-a",
"-b", or "-c" under imediff(diff3) may be useful. (TBH, I don't know how
useful these are.)

Customization TUI
=================

The imediff program can customize its key command binding and its color
setting using the "~/.imediff" file in the ini file format.  You can create
its template file by the "imediff -t" command.  If this file is missing,
default settings are used.

The left side of this configuration file is the keys described in the above.
The right side is your configuration choices.  The current settings can be
confirmed by the "/" key dialog on the interactive TUI.

You can disable the existing "~/.imediff" file without renaming it by
specifying "none" as "imediff -C none ...". Then, only the internal default
values of the imediff program are used.

Key MACRO
=========

Automatic processing with imediff can be enabled using key MACRO featurs
with the "-M" option.

For TUI mode, TUI is started after processing the MACRO.

For CLI mode, you must add "w" at the end of to the MACRO command to write
the result to a file.  For example, "imediff -n -MAbw ...".

Log file
========

After version 3.4.0, imediff doesn't create log file as its default
behavior.

When "--force-logging" option is specified, imediff forces to generate
"imediff.log" at the current directory as the log file with its log level at
"INFO".

You may explicitly set the log file and its log level with "--logfile" and
"--loglevel" options.

The "git-ime" command invokes "imediff --logfile=.git/imediff.log ..." to
generate the log file in the "./git/" directory and avoid interfarence with
the git repository.

Internal logic
==============

Here is a brief overview of the internal logic of imediff.  Normally, you
don't need to tweak it.

The imediff command internally uses "difflib.SequenceMatch" class provided
by the Python Standard Library.  The sequence may be a sequence of string
for line matching or a sequence of character for character matching
depending on its usage point.  "isjunk" parameter for the SequenceMatch
instance may be tweaked using "--isjunk" option.

The imediff tries its best to match lines using 2 step approach.

 * Step 1: full line match on lines after removing whitespaces etc.
 * Step 2: shortened line match for non-identical lines (head or tail).

Step 1 can be tweaked using the "--line-rule=N" option to set the line
filtering rule.

 * N=0:  strip leading and tailing whitespaces
 * N=1:  strip all whitespaces
 * N=2:  strip all whitespaces and quotation marks (default)
 * N=3:  strip all non-alphanumerics
 * N=10: strip leading and tailing whitespaces and lowecase all characters
 * N=11: strip all whitespaces and lowecase all characters
 * N=12: strip all whitespaces and quotation marks and lowecase all
         characters
 * N=13: strip all non-alphanumerics and lowecase all characters

Step 2 can be tweaked using:

 * "--line-min LINE_MIN" option to set the minimum partial line match length
 * "--line-max LINE_MAX" option to set the maximum partial line match length
 * "--line-factor LINE_FACTOR" option to set the shortening factor (1-9) for
   the partial line match length.  default=8 meaning 80% per step

git-ime
========

The git-ime command is a companion script provided with imediff.  It is a
wrapper tool for git and imediff to disect a single commit containing many
changes.

When a commit contains changes to multiple files, it splits the commit into
many commits of single file changes.

When a commit contains changes to a single file, it splits the commit into
many minimum commits using "imediff -n -MAbw ..." repeatedly.

Use of git-ime and imediff along with "git rebase -i ..." can clean
intertwined change history.  This may be useful for debugging.

Note
====

The version 3.0 is a major rewrite to address file size limitation caused by
the underlying curses library and to limit wdiff to operate only within a
single line.

The version 3.4.0 changes logging behavior and redesigns short command
options for imediff.

The "diff3 -m file_a file_b file_c" has an odd feature of showing diff2
between "file_b" and "file_c" for the portion of changes in which both
"file_a" and "file_c" underwent identical changes from "file_b". This
imediff program results is a more intuitive one with the clean merge.  If
you are after such identical changes, you may use "-b" option with imediff
(diff3).""".format(
    __version__
)


class TextPad(TextData):  # TUI data
    """Curses class to handle diff data for 2 or 3 lines"""

    ####################################################################
    # Externally exposed initializer method
    ####################################################################
    #  self.chunk_list[chunk_index] = (                # for CLI and TUI
    #      tag,             # diff opcode tag EeNFAC
    #      i1,              # self.list_a (start)-index
    #      i2,              # self.list_a (end+1)-index
    #      j1,              # self.list_b (start)-index
    #      j2,              # self.list_b (end+1)-index
    #      k1,              # self.list_c (start)-index
    #      k2,              # self.list_c (end+1)-index
    #      action,      # data action abcdefg (self.default_action)
    #      merge_buffer,    # merge_buffer for editor ([])
    #  ) # chunk_list item tuple (9 param)
    #
    #  self.usr_chunk_list: list of user accessible chunk_index
    #
    #  self.chunk_to_virt[chunk_index]: virt_row (TUI)
    #
    def __init__(self, list_a, list_b, list_c, args, confs):
        # Init from super class "TextData"
        super().__init__(list_a, list_b, list_c, args, confs)
        logger.debug("starting ...")
        self.init_args_confs_tui(args, confs)
        self.chunk_to_virt = list()
        self.virt_to_chunk = list()
        logger.debug("finished")
        return

    def init_args_confs_tui(self, args, confs):
        self.attrib = confs["attrib"]
        # self.poke = args.poke

    ####################################################################
    # Externally used main method and effective main method
    ####################################################################
    # Design concept:
    #     * Use full terminal screen
    #     * Access them directly (no window nor pad)
    #     * String operation by pointers
    #
    # See https://docs.python.org/3/library/curses.html
    #     https://docs.python.org/3/howto/curses.html
    #
    # curses terminal screen absolute coordinate as (y,x):
    # +----------------------------------------------------------------+
    # | (0,0) top-left corner                        (stdscr_col_max-1)|
    # |  +--> x-axis                                                   |
    # |  |                                                             |
    # |  |                    TEXT DATA DISPLAY                        |
    # |  V y-axis                                                      |
    # |                                                                |
    # |                                                                |
    # | (stdscr_row_max-2,0)        (stdscr_row_max-2,stdscr_col_max-1)|
    # +----------------------------------------------------------------+
    # | (stdscr_row_max-1,0) STATUS (stdscr_row_max-1,stdscr_col_max-1)|
    # +----------------------------------------------------------------+
    #
    # Here:
    #     * stdscr_row_max, stdscr_col_max = stdscr.getmaxyx()
    #
    # Internal data:
    #   * CLI
    #     * chunk_list[chunk_index]: -> chunk_list item tuple
    #     * usr_chunk_list[usr_chunk_index]: -> chunk_index
    #   * TUI
    #     * chunk_to_virt[chunk_index] = virt_row
    #     * virt_to_row[virt_row] = (chunk_index, chunk_subindex, action)
    # Terminal size: 80 col x 24 row required
    #
    ####################################################################

    def main(self):  # for curses TUI (wrapper)
        """Interactive driven by CURSES-TUI"""
        logger.debug("start with macro = '{}'".format(self.macro))
        curses.wrapper(self.tui_loop)
        logger.debug("finished")
        return

    def tui_loop(self, stdscr):  # for curses TUI (core)
        logger.debug("start with macro = '{}'".format(self.macro))
        # initialize
        curses.curs_set(0)  # cursor off
        self.init_curses(stdscr)
        # display parameters
        flag_update_corner = True
        corner_virt_row = 0
        corner_virt_col = 0
        while True:
            self.remap_chunk_virt()
            if len(self.chunk_list) != len(self.chunk_to_virt):
                logger.error(
                    "E: insane: len(chunk_list) != len(chunk_to_virt): {} {}".format(
                        len(self.chunk_list), len(self.virt_to_chunk)
                    )
                )
                sys.exit(2)
            if flag_update_corner:
                if self.focused_usr_chunk_index is None:
                    logger.info(
                        "W: focused_usr_chunk_index=None (cleanly merged files?)"
                    )
                    corner_virt_row = 0
                else:
                    focused_chunk_index = self.get_chunk_index_from_usr_chunk_list(
                        self.focused_usr_chunk_index
                    )
                    if focused_chunk_index is not None:
                        stdscr_row_max, _ = self.stdscr.getmaxyx()
                        focused_virt_row = self.chunk_to_virt[focused_chunk_index]
                        corner_virt_row = max(
                            focused_virt_row - ((stdscr_row_max - 1) // 3), 0
                        )
                        logger.debug(
                            "focused_chunk_index={} >> focused_virt_row={} >> corner_virt_row={}".format(
                                focused_chunk_index, focused_virt_row, corner_virt_row
                            )
                        )
            flag_update_corner = False
            self.display_data(corner_virt_row, corner_virt_col)
            keyname = self.get_macro_command()
            if keyname in ["IGNORE"]:
                logger.warning("W: ignore key/macro input")
                pass
            elif keyname in ["w", "x"]:
                if self.sloppy or (
                    not self.sloppy and self.get_unresolved_count() == 0
                ):
                    if not self.confirm_exit or self.display_content_win(
                        [
                            (
                                "Do you want to 'save and exit'?",
                                self.get_attr("color_warn", False),
                            ),
                            ("", self.get_attr("color_white", False)),
                            (
                                "* Press '{y:}' to save and exit.".format(
                                    y=self.rkc["y"]
                                ),
                                self.get_attr("color_white_bold", False),
                            ),
                            (
                                "* Press '{SPACE:}' to continue.".format(
                                    SPACE=self.rkc["SPACE"]
                                ),
                                self.get_attr("color_white_bold", False),
                            ),
                        ],
                        ["y", "Y", "N", "n", "SPACE", "ESCAPE"],
                    ) in ["y", "Y"]:
                        write_file(self.file_o, self.get_string_from_content_for_file())
                        break
                else:
                    self.display_content_win(
                        [
                            (
                                "You can't 'save and exit' due to the non-clean merge.",
                                self.get_attr("color_warn", False),
                            ),
                            ("", self.get_attr("color_white", False)),
                            (
                                "* Press '{SPACE:}' to continue.".format(
                                    SPACE=self.rkc["SPACE"]
                                ),
                                self.get_attr("color_white_bold", False),
                            ),
                            ("", self.get_attr("color_white", False)),
                            (
                                "You can evade this requirement by pressing 'm' on all non-clean",
                                self.get_attr("color_white", False),
                            ),
                            (
                                "merges and making them as manually edited merged data.",
                                self.get_attr("color_white", False),
                            ),
                            ("", self.get_attr("color_white", False)),
                            (
                                "This requirement can be disabled by starting this program as",
                                self.get_attr("color_white", False),
                            ),
                            (
                                "'imediff --sloppy ...', too.",
                                self.get_attr("color_white", False),
                            ),
                        ],
                        ["SPACE", "y", "Y", "w", "W", "e", "E"],
                    )
                    # return value is ignored
            elif keyname in ["QUIT", "q", "Q"]:
                if not self.confirm_exit or self.display_content_win(
                    [
                        (
                            "Do you want to 'quit without saving'?",
                            self.get_attr("color_warn", False),
                        ),
                        ("", self.get_attr("color_white", False)),
                        (
                            "* Press '{y:}' to quit without saving.".format(
                                y=self.rkc["y"]
                            ),
                            self.get_attr("color_white_bold", False),
                        ),
                        (
                            "* Press '{SPACE:}' to continue.".format(
                                SPACE=self.rkc["SPACE"]
                            ),
                            self.get_attr("color_white_bold", False),
                        ),
                    ],
                    ["y", "Y", "N", "n", "q", "Q", "SPACE", "ESCAPE"],
                ) in ["y", "Y"]:
                    # only "y" or "Y" quit.  "q", "Q", ... are ignored
                    self.chunk_list = []
                    logger.error("Quit without saving by the user request")
                    sys.exit(2)
            elif keyname in ["?", "/", "F1"]:
                # Show help screen (short so exit with SPACE)
                self.display_content_win(
                    self.get_helptext(), ["SPACE", "ESCAPE", "q", "Q"]
                )
            elif keyname == "t":
                # Show tutorial screen (long so no-exit with SPACE)
                self.display_popup_win(tutorial, ["ESCAPE", "q", "Q"], "color_white")
            # Moves in document
            elif keyname in ["j", "DOWN"]:
                corner_virt_row += 1
            elif keyname in ["k", "UP"]:
                corner_virt_row -= 1
            elif keyname in ["l", "RIGHT"]:
                corner_virt_col += 8
            elif keyname in ["h", "LEFT"]:
                corner_virt_col -= 8
            elif keyname in ["PAGEDOWN"]:
                corner_virt_row += 20
            elif keyname in ["PAGEUP"]:
                corner_virt_row -= 20
            # Terminal resize signal
            else:
                pass
            # Following key-command updates TextPad
            if self.focused_usr_chunk_index is not None:
                chunk_index = self.usr_chunk_list[self.focused_usr_chunk_index]
                # record current action
                action = self.get_action(chunk_index)
                # Explicitly select chunk action
                if keyname in ["a", "b", "c", "d", "e", "f", "g"]:
                    self.set_action(chunk_index, keyname)
                    flag_update_corner = True
                elif keyname in ["1", "2", "3", "4", "5", "6", "7"]:
                    self.set_action(
                        chunk_index,
                        chr(ord(keyname) - ord("1") + ord("a")),
                    )
                    flag_update_corner = True
                elif keyname in ["", "A", "B", "C", "D", "E", "F", "G"]:
                    self.set_action_all(keyname.lower())
                    flag_update_corner = True
                elif keyname in ["ENTER"]:
                    # Enter key will rotate action setting
                    # 2: a -> b      -> d -> [e ->] f -> a
                    # 3: a -> b -> c -> d -> [e ->] f -> a
                    if action == "a":
                        self.set_action(chunk_index, "b")
                    elif action == "b" and self.diff_mode == 2:
                        self.set_action(chunk_index, "d")
                    elif action == "b" and self.diff_mode == 3:
                        self.set_action(chunk_index, "c")
                    elif action == "c":
                        self.set_action(chunk_index, "d")
                    elif (
                        action == "d" and self.get_merge_buffer(chunk_index) is not None
                    ):
                        self.set_action(chunk_index, "e")
                    elif action == "d":
                        self.set_action(chunk_index, "f")
                    elif action == "e":
                        self.set_action(chunk_index, "f")
                    else:  # f
                        self.set_action(chunk_index, "a")
                    flag_update_corner = True
                elif keyname in ["m"]:
                    self.set_updated_merge_buffer(chunk_index)
                    flag_update_corner = True
                elif keyname in ["M"]:
                    self.set_deleted_merge_buffer(chunk_index)
                    flag_update_corner = True
                elif keyname in ["n", "SPACE"]:
                    self.move_focus_to_any_resolvable_chunk_next()
                    flag_update_corner = True
                elif keyname in ["p", "BACKSPACE"]:
                    self.move_focus_to_any_resolvable_chunk_prev()
                    flag_update_corner = True
                elif keyname in ["0", "HOME"]:
                    self.move_focus_to_any_resolvable_chunk_home()
                    flag_update_corner = True
                elif keyname in ["9", "END"]:
                    self.move_focus_to_any_resolvable_chunk_end()
                    flag_update_corner = True
                elif keyname in ["N", "TAB"]:
                    self.move_focus_to_usr_chunk_next()
                    flag_update_corner = True
                elif keyname in ["P", "BTAB"]:
                    self.move_focus_to_usr_chunk_prev()
                    flag_update_corner = True
                elif keyname in [")"]:
                    self.move_focus_to_usr_chunk_home()
                    flag_update_corner = True
                elif keyname in ["("]:
                    self.move_focus_to_usr_chunk_end()
                    flag_update_corner = True
                else:
                    pass
            # lower boundary check
            corner_virt_col = max(0, corner_virt_col)
            corner_virt_row = max(0, corner_virt_row)
            # TODO: upper boundary check
        logger.debug("finished")
        return

    ####################################################################
    # Internally used utility methods (update internal data)
    ####################################################################
    # This is brute force update of all the internal data
    # This may be updated to update only data affected

    def remap_chunk_virt(self):
        virt_row_next = 0
        self.chunk_to_virt = list()
        # [ virt_row, ...]
        self.virt_to_chunk = list()
        # [(chunk_index, chunk_subindex, action), ...]
        for chunk_index in range(len(self.chunk_list)):
            virt_row = virt_row_next
            self.chunk_to_virt.append(virt_row)
            (
                tag,
                i1,
                i2,
                j1,
                j2,
                k1,
                k2,
                action,
                merge_buffer,
            ) = self.chunk_list[chunk_index]
            # virt_range
            if action == "=" or action == "#":
                virt_range = i2 - i1
                if virt_range > 0:
                    for virt_row_index in range(virt_row, virt_row + virt_range):
                        chunk_subindex = virt_row_index - virt_row
                        self.virt_to_chunk.append((chunk_index, chunk_subindex, action))
                else:
                    # no content consumes 1 line for "???"
                    virt_range = 1
                    self.virt_to_chunk.append((chunk_index, 0, action))
            elif action == "a" or action == "A":
                virt_range = i2 - i1
                if virt_range > 0:
                    for virt_row_index in range(virt_row, virt_row + virt_range):
                        chunk_subindex = virt_row_index - virt_row
                        self.virt_to_chunk.append((chunk_index, chunk_subindex, action))
                else:
                    # no content consumes 1 line for "???"
                    virt_range = 1
                    self.virt_to_chunk.append((chunk_index, 0, action))
            elif action == "b" or action == "B":
                virt_range = j2 - j1
                if virt_range > 0:
                    for virt_row_index in range(virt_row, virt_row + virt_range):
                        chunk_subindex = virt_row_index - virt_row
                        self.virt_to_chunk.append((chunk_index, chunk_subindex, action))
                else:
                    # no content consumes 1 line for "???"
                    virt_range = 1
                    self.virt_to_chunk.append((chunk_index, 0, action))
            elif action == "c" or action == "C":
                virt_range = k2 - k1
                if virt_range > 0:
                    for virt_row_index in range(virt_row, virt_row + virt_range):
                        chunk_subindex = virt_row_index - virt_row
                        self.virt_to_chunk.append((chunk_index, chunk_subindex, action))
                else:
                    # no content consumes 1 line for "???"
                    virt_range = 1
                    self.virt_to_chunk.append((chunk_index, 0, action))
            elif action == "d" and self.diff_mode == 2:
                self.virt_to_chunk.append((chunk_index, 0, "d20"))
                virt_range_a = i2 - i1
                for virt_row_index in range(virt_row, virt_row + virt_range_a):
                    chunk_subindex = virt_row_index - virt_row
                    self.virt_to_chunk.append((chunk_index, chunk_subindex, "a"))
                self.virt_to_chunk.append((chunk_index, 0, "d21"))
                virt_range_b = j2 - j1
                for virt_row_index in range(virt_row, virt_row + virt_range_b):
                    chunk_subindex = virt_row_index - virt_row
                    self.virt_to_chunk.append((chunk_index, chunk_subindex, "b"))
                self.virt_to_chunk.append((chunk_index, 0, "d22"))
                # diff2 consumes 3 extra lines as separators
                virt_range = virt_range_a + virt_range_b + 3
            elif action == "d" and self.diff_mode == 3:
                self.virt_to_chunk.append((chunk_index, 0, "d30"))
                virt_range_a = i2 - i1
                for virt_row_index in range(virt_row, virt_row + virt_range_a):
                    chunk_subindex = virt_row_index - virt_row
                    self.virt_to_chunk.append((chunk_index, chunk_subindex, "a"))
                self.virt_to_chunk.append((chunk_index, 0, "d31"))
                virt_range_b = j2 - j1
                for virt_row_index in range(virt_row, virt_row + virt_range_b):
                    chunk_subindex = virt_row_index - virt_row
                    self.virt_to_chunk.append((chunk_index, chunk_subindex, "b"))
                self.virt_to_chunk.append((chunk_index, 0, "d32"))
                virt_range_c = k2 - k1
                for virt_row_index in range(virt_row, virt_row + virt_range_c):
                    chunk_subindex = virt_row_index - virt_row
                    self.virt_to_chunk.append((chunk_index, chunk_subindex, "c"))
                self.virt_to_chunk.append((chunk_index, 0, "d33"))
                # diff2 consumes 3 extra lines as separators
                virt_range = virt_range_a + virt_range_b + virt_range_c + 4
            elif (action == "e" or action == "G") and len(merge_buffer) > 0:
                virt_range = len(merge_buffer)
                for chunk_subindex in range(virt_range):
                    self.virt_to_chunk.append((chunk_index, chunk_subindex, action))
            elif (
                action == "f" and i2 - i1 == 1 and j2 - j1 == 1 and self.diff_mode == 2
            ) or (
                action == "f"
                and i2 - i1 == 1
                and j2 - j1 == 1
                and k2 - k1 == 1
                and self.diff_mode == 3
            ):
                virt_range = 1
                self.virt_to_chunk.append((chunk_index, 0, "f"))
            else:
                logger.error(
                    "E: bad combination - diff{} action: {}, tag: {}, a[{}:{}] b[{}:{}] c[{}:{}] len[e]={}".format(
                        self.diff_mode,
                        action,
                        tag,
                        i1,
                        i2,
                        j1,
                        j2,
                        k1,
                        k2,
                        len(merge_buffer),
                    )
                )
                sys.exit(2)
            # This increments virt_row for virt_range per loop
            virt_row_next = virt_row + virt_range
        # debug
        for chunk_index, virt_row in enumerate(self.chunk_to_virt):
            logger.debug("chunk[{}] --> virt_row[{}]".format(chunk_index, virt_row))

        for virt_row, (chunk_index, chunk_subindex, action) in enumerate(
            self.virt_to_chunk
        ):
            logger.debug(
                "virt_row[{}] --> (chunk[{}], chunk_subindex={}, action:{})".format(
                    virt_row, chunk_index, chunk_subindex, action
                )
            )
        logger.debug(
            "len(chunk_list)={}, len(usr_chunk_list)={}, len(virt_to_chunk)={}, len(chunk_to_virt)={}".format(
                len(self.chunk_list),
                len(self.usr_chunk_list),
                len(self.virt_to_chunk),
                len(self.chunk_to_virt),
            )
        )
        return

    ####################################################################
    # Internally used utility methods (initializer within tui_main)
    ####################################################################
    def init_curses(self, stdscr):
        self.stdscr = stdscr
        stdscr_row_max, stdscr_col_max = self.stdscr.getmaxyx()
        if stdscr_row_max < 20 or stdscr_col_max < 60:
            logger.error(
                "E: terminal size too small: {} {}".format(
                    stdscr_row_max, stdscr_col_max
                )
            )
            sys.exit(2)
        self.stdscr.clear()
        self.stdscr.refresh()
        # set up color
        curses.start_color()

        # C-library
        # COLOR_BLACK   0
        # COLOR_RED     1
        # COLOR_GREEN   2
        # COLOR_YELLOW  3
        # COLOR_BLUE    4
        # COLOR_MAGENTA 5
        # COLOR_CYAN    6
        # COLOR_WHITE   7

        # curses.init_pair(0, curses.COLOR_WHITE,   curses.COLOR_BLACK)
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(4, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
        curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(7, curses.COLOR_RED, curses.COLOR_WHITE)
        curses.init_pair(8, curses.COLOR_GREEN, curses.COLOR_WHITE)
        curses.init_pair(9, curses.COLOR_YELLOW, curses.COLOR_WHITE)
        curses.init_pair(10, curses.COLOR_BLUE, curses.COLOR_WHITE)
        curses.init_pair(11, curses.COLOR_MAGENTA, curses.COLOR_WHITE)
        curses.init_pair(12, curses.COLOR_CYAN, curses.COLOR_WHITE)

        self.curses_value = dict()
        self.curses_value["WHITE"] = curses.color_pair(0)
        self.curses_value["RED"] = curses.color_pair(1)
        self.curses_value["GREEN"] = curses.color_pair(2)
        self.curses_value["YELLOW"] = curses.color_pair(3)
        self.curses_value["BLUE"] = curses.color_pair(4)
        self.curses_value["MAGENTA"] = curses.color_pair(5)
        self.curses_value["CYAN"] = curses.color_pair(6)
        self.curses_value["RED/WHITE"] = curses.color_pair(7)
        self.curses_value["GREEN/WHITE"] = curses.color_pair(8)
        self.curses_value["YELLOW/WHITE"] = curses.color_pair(9)
        self.curses_value["BLUE/WHITE"] = curses.color_pair(10)
        self.curses_value["MAGENTA/WHITE"] = curses.color_pair(11)
        self.curses_value["CYAN/WHITE"] = curses.color_pair(12)
        self.curses_value["DIM"] = curses.A_DIM
        self.curses_value["NORMAL"] = curses.A_NORMAL
        self.curses_value["BOLD"] = curses.A_BOLD
        self.curses_value["REVERSE"] = curses.A_REVERSE

        if not curses.has_colors():
            self.curses_value["WHITE"] = curses.color_pair(0)
            self.curses_value["RED"] = curses.color_pair(0)
            self.curses_value["GREEN"] = curses.color_pair(0)
            self.curses_value["YELLOW"] = curses.color_pair(0)
            self.curses_value["BLUE"] = curses.color_pair(0)
            self.curses_value["MAGENTA"] = curses.color_pair(0)
            self.curses_value["CYAN"] = curses.color_pair(0)
            self.curses_value["RED/WHITE"] = curses.color_pair(0)
            self.curses_value["GREEN/WHITE"] = curses.color_pair(0)
            self.curses_value["YELLOW/WHITE"] = curses.color_pair(0)
            self.curses_value["BLUE/WHITE"] = curses.color_pair(0)
            self.curses_value["MAGENTA/WHITE"] = curses.color_pair(0)
            self.curses_value["CYAN/WHITE"] = curses.color_pair(0)

    ####################################################################
    # Color
    ####################################################################

    def get_attr(self, data_type, focus):
        #
        if focus:
            attrib_key = data_type + "_" + "focus"
        else:
            attrib_key = data_type
        value = self.attrib.get(attrib_key, "WHITE,NORMAL").split(",")
        attrib = 0
        for a in value:
            attrib |= self.curses_value[a]
        return attrib

    def get_color(self, data_type):
        color = self.attrib.get(data_type, "WHITE,NORMAL").split(",")[0]
        return color

    def get_macro_command(self):  # overriding for TUI
        """Macro parsing instead of curses getch"""
        if len(self.macro) == 0:
            keyname = ""  # end of MACRO and exit
        else:
            keyname = self.macro[:1]
            self.macro = self.macro[1:]
            if keyname == "[":
                pos = self.macro.find("]")
                if pos >= 0:
                    # look for ]
                    keyname = self.macro[:pos]
                    self.macro = self.macro[1 + pos :]
        if keyname == "":  # interactive
            try:
                keyname = get_keyname(self.stdscr.getch())
            except Exception as _:
                keyname = ""  # quit w/o saving for ^C
            keyname = self.kc.get(keyname, "IGNORE")
        logger.debug("key={}".format(keyname))
        return keyname

    def get_helptext(self):
        if self.diff_mode == 2:
            content = [
                (
                    "    Interactive Merge Editor to merge 2 DIFFerent files",
                    self.get_attr("color_white_bold", False),
                ),
                ("", self.get_attr("color_white", False)),
                (
                    "imediff processes 2 different input files using one of 5 actions:",
                    self.get_attr("color_white", False),
                ),
                (
                    " * action a: display {file_a} (older) in {color_a}".format(
                        file_a=self.file_a, color_a=self.get_color("color_a")
                    ),
                    self.get_attr("color_a", False),
                ),
                (
                    " * action b: display {file_b} (newer) in {color_b}".format(
                        file_b=self.file_b,
                        color_b=self.get_color("color_b2"),
                    ),
                    self.get_attr("color_b2", False),
                ),
                (
                    " * action d: display diff2(a,b) with marker",
                    self.get_attr("color_diff_marker", False),
                ),
                (
                    " * action e: display editor result buffer in {color_editor}".format(
                        color_editor=self.get_color("color_editor")
                    ),
                    self.get_attr("color_editor", False),
                ),
                (
                    " * action f: display wdiff2(a,b) with marker",
                    self.get_attr("color_wdiff_marker", False),
                ),
                ("", self.get_attr("color_white", False)),
            ]
            for item in """\
key commands          induced actions
{w},{x}                   write and exit
{q}                     quit without saving
{a}/{b}/{d}/{e}/{f}             set a chunk to a/b/d/e/f action
1/2/4/5/6             set a chunk to a/b/d/e/f action (alternative)
{A}/{B}/{D}/{E}/{F}             set all chunks to a/b/d/e/f action
enter                 toggle action of a chunk
{m}                     modify a chunk with editor: {edit_cmd}
{M}                     remove a editor result buffer
arrows/pgdn,{j}/pgup,{k}  move scope of the display
space,{n} /backspace,{p}  select the next/previous usr_chunk
tab,{N}   /shift-tab,{P}  select the next/previous unresolved usr_chunk
{zero}       /{nine}            select the first/last usr_chunk
{rparen},home  /{lparen},end        select the first/last unresolved usr_chunk
{question},{slash}                   show this help
{t}                     show tutorial""".format(
                w=self.rkc["w"],
                x=self.rkc["x"],
                q=self.rkc["q"],
                a=self.rkc["a"],
                b=self.rkc["b"],
                d=self.rkc["d"],
                e=self.rkc["e"],
                f=self.rkc["f"],
                A=self.rkc["A"],
                B=self.rkc["B"],
                D=self.rkc["D"],
                E=self.rkc["E"],
                F=self.rkc["F"],
                m=self.rkc["m"],
                edit_cmd=self.edit_cmd,
                M=self.rkc["M"],
                j=self.rkc["j"],
                k=self.rkc["k"],
                n=self.rkc["n"],
                p=self.rkc["p"],
                N=self.rkc["N"],
                P=self.rkc["P"],
                zero=self.rkc["0"],
                nine=self.rkc["9"],
                lparen=self.rkc["("],
                rparen=self.rkc[")"],
                t=self.rkc["t"],
                question=self.rkc["?"],
                slash=self.rkc["/"],
            ).split(
                "\n"
            ):
                content.append((item, self.get_attr("color_white", False)))
        else:
            content = [
                (
                    "    Interactive Merge Editor to merge 3 DIFFerent files",
                    self.get_attr("color_white_bold", False),
                ),
                ("", self.get_attr("color_white", False)),
                (
                    "imediff processes 3 different input files using one of 7 actions:",
                    self.get_attr("color_white", False),
                ),
                (
                    " * action a: display {file_a} (MYFILE) in {color_a}".format(
                        file_a=self.file_a, color_a=self.get_color("color_a")
                    ),
                    self.get_attr("color_a", False),
                ),
                (
                    " * action b: display {file_b} (OLDER) in {color_b}".format(
                        file_b=self.file_b,
                        color_b=self.get_color("color_b3"),
                    ),
                    self.get_attr("color_b3", False),
                ),
                (
                    " * action c: display {file_c} (YOURFILE) in {color_c}".format(
                        file_c=self.file_c,
                        color_c=self.get_color("color_c"),
                    ),
                    self.get_attr("color_c", False),
                ),
                (
                    " * action d: display diff3(a,b,c) with marker",
                    self.get_attr("color_diff_marker", False),
                ),
                (
                    " * action e: display editor result buffer in {color_editor}".format(
                        color_editor=self.get_color("color_editor")
                    ),
                    self.get_attr("color_editor", False),
                ),
                (
                    " * action f: display wdiff3(a,b,c) with marker",
                    self.get_attr("color_wdiff_marker", False),
                ),
                ("", self.get_attr("color_white", False)),
            ]
            for item in """\
key commands          induced actions
{w},{x}                   write and exit
{q}                     quit without saving
{a}/{b}/{c}/{d}/{e}/{f}/{g}         set a chunk to a/b/c/d/e/f/g action
1/2/3/4/5/6/7         set a chunk to a/b/c/d/e/f/g action (alternative)
{A}/{B}/{C}/{D}/{E}/{F}/{G}         set all chunks to a/b/c/d/e/f/g action
enter                 toggle action of a chunk
{m}                     modify a chunk with editor: {edit_cmd}
{M}                     remove a editor result buffer
arrows/pgdn,{j}/pgup,{k}  move scope of the display
space,{n} /backspace,{p}  select the next/previous usr_chunk
tab,{N}   /shift-tab,{P}  select the next/previous unresolved usr_chunk
{zero}       /{nine}            select the first/last usr_chunk
{rparen},home  /{lparen},end        select the first/last unresolved usr_chunk
{question},{slash}                   show this help
{t}                     show tutorial""".format(
                w=self.rkc["w"],
                x=self.rkc["x"],
                q=self.rkc["q"],
                a=self.rkc["a"],
                b=self.rkc["b"],
                c=self.rkc["c"],
                d=self.rkc["d"],
                e=self.rkc["e"],
                f=self.rkc["f"],
                g=self.rkc["g"],
                A=self.rkc["A"],
                B=self.rkc["B"],
                C=self.rkc["C"],
                D=self.rkc["D"],
                E=self.rkc["E"],
                F=self.rkc["F"],
                G=self.rkc["G"],
                m=self.rkc["m"],
                edit_cmd=self.edit_cmd,
                M=self.rkc["M"],
                j=self.rkc["j"],
                k=self.rkc["k"],
                n=self.rkc["n"],
                p=self.rkc["p"],
                N=self.rkc["N"],
                P=self.rkc["P"],
                zero=self.rkc["0"],
                nine=self.rkc["9"],
                lparen=self.rkc["("],
                rparen=self.rkc[")"],
                t=self.rkc["t"],
                question=self.rkc["?"],
                slash=self.rkc["/"],
            ).split(
                "\n"
            ):
                content.append((item, self.get_attr("color_white", False)))
        return content

    ####################################################################
    # focus off
    # action["="]       "color_merge_abc"        # diff3
    # action["="]       "color_merge_ab"         # diff2
    # action["#"]       "color_merge_ac"         # diff3
    # action["A"]       "color_merge_a"          # diff3
    # action["C"]       "color_merge_c"          # diff3
    # action["G"]       "color_merge_wdiff"      # diff3
    # focus on/off
    # action["a"]       "color_a"                # diff2, diff3
    # action["b"]       "color_b2"               # diff2
    # action["b"]       "color_b3"               # diff3
    # action["c"]       "color_c"                # diff3
    # action["e"]       "color_editor"           # diif2, diff3
    # action["d"]       "color_diff_marker"             # diff2, diff3
    # action["f"]       "color_wdiff"            # diff2, diff3
    # TBD
    #                   "color_marker"
    #                   "color_status"
    #                   "color_type"
    #                   "color_zero"

    def display_data(self, corner_virt_row, corner_virt_col):
        stdscr_row_max, stdscr_col_max = self.stdscr.getmaxyx()
        # text_data 0 ................... < stdscr_row_max - 1
        # stat_data stdscr_row_max -1 ... < stdscr_row_max
        for row_index in range(stdscr_row_max - 1):
            virt_row_index = row_index + corner_virt_row
            if virt_row_index < len(self.virt_to_chunk):
                chunk_index, chunk_subindex, action = self.virt_to_chunk[virt_row_index]
                logger.debug(
                    "virt_row_index={} row_index={} chunk_index={} chunk_subindex={} action:{}".format(
                        virt_row_index, row_index, chunk_index, chunk_subindex, action
                    )
                )
                (
                    tag,  # tag
                    i1,
                    i2,
                    j1,
                    j2,
                    k1,
                    k2,
                    _,  # action
                    merge_buffer,
                ) = self.chunk_list[chunk_index]
                #
                if self.focused_usr_chunk_index is None:
                    focus = False
                elif chunk_index == self.get_chunk_index_from_usr_chunk_list(
                    self.focused_usr_chunk_index
                ):
                    focus = True
                else:
                    focus = False
                if action == "=" and self.diff_mode == 2:
                    logger.debug(
                        "action:{} row_index={} with i1={}".format(
                            action, row_index, i1
                        )
                    )
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.list_a[i1],
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_merge_ab", focus),
                            )
                        ],
                        action,
                    )
                #
                elif action == "=" and self.diff_mode == 3:
                    if i1 < i2:
                        logger.debug(
                            "action:{} row_index={} with i1={}".format(
                                action, row_index, i1
                            )
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_a[i1],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_abc", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with i1={} i2={}".format(row_index, i1, i2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_a", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "#" and self.diff_mode == 3:
                    if i1 < i2:
                        logger.debug(
                            "action:{} row_index={} with i1={}".format(
                                action, row_index, i1
                            )
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_a[i1],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_ac", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with i1={} i2={}".format(row_index, i1, i2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_a", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "A":
                    if i1 < i2:
                        logger.debug(
                            "row_index={} with i1={} i2={}".format(row_index, i1, i2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_a[i1 + chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_a", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with i1={} i2={}".format(row_index, i1, i2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_a", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "C" and self.diff_mode == 3:
                    if k1 < k2:
                        logger.debug(
                            "row_index={} with k1={} k2={}".format(row_index, k1, k2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_c[k1 + chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_c", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with k1={} k2={}".format(row_index, k1, k2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_c", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "G" and self.diff_mode == 3:
                    if len(merge_buffer) > 0:
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    merge_buffer[chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_wdiff", focus),
                                )
                            ],
                            action,
                            clrtoeol=True,
                        )
                    else:  #
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_merge_wdiff", focus),
                                )
                            ],
                            action,
                            clrtoeol=True,
                        )
                #
                elif action == "a":
                    if i1 < i2:
                        logger.debug(
                            "row_index={} with i1={} i2={}".format(row_index, i1, i2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_a[i1 + chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_a", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with i1={} i2={}".format(row_index, i1, i2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_a", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "b" and self.diff_mode == 2:  #
                    if j1 < j2:
                        logger.debug(
                            "row_index={} with j1={} j2={}".format(row_index, j1, j2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_b[j1 + chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_b2", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with j1={} j2={}".format(row_index, j1, j2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_b2", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "b" and self.diff_mode == 3:  #
                    if j1 < j2:
                        logger.debug(
                            "row_index={} with j1={} j2={}".format(row_index, j1, j2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_b[j1 + chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_b3", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with j1={} j2={}".format(row_index, j1, j2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_b3", focus),
                                )
                            ],
                            action,
                        )
                #
                elif action == "c" and self.diff_mode == 3:
                    if k1 < k2:
                        logger.debug(
                            "row_index={} with k1={} k2={}".format(row_index, k1, k2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    self.list_c[k1 + chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_c", focus),
                                )
                            ],
                            action,
                        )
                    else:  # i1 == i2
                        logger.debug(
                            "row_index={} with k1={} k2={}".format(row_index, k1, k2)
                        )
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_c", focus),
                                )
                            ],
                            "c",
                        )
                #
                elif action == "e":
                    if len(merge_buffer) > 0:
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    merge_buffer[chunk_subindex],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_editor", focus),
                                )
                            ],
                            action,
                            clrtoeol=True,
                        )
                    else:  #
                        self.display_imediff_content(
                            self.stdscr,
                            row_index,
                            [
                                (
                                    "??? (" + action + ")",
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_editor", focus),
                                )
                            ],
                            action,
                            clrtoeol=True,
                        )
                #
                elif action == "d20" and self.diff_mode == 2:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls0 % self.file_a,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                elif action == "d21" and self.diff_mode == 2:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls2,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                elif action == "d22" and self.diff_mode == 2:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls3 % self.file_b,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                #
                elif action == "d30" and self.diff_mode == 3:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls0 % self.file_a,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                elif action == "d31" and self.diff_mode == 3:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls1 % self.file_b,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                elif action == "d32" and self.diff_mode == 3:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls2,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                elif action == "d33" and self.diff_mode == 3:
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        [
                            (
                                self.ls3 % self.file_c,
                                corner_virt_col,
                                corner_virt_col + stdscr_col_max,
                                self.get_attr("color_diff_marker", focus),
                            )
                        ],
                        action,
                        clrtoeol=True,
                    )
                #
                elif (
                    action == "f"
                    and self.diff_mode == 2
                    and i2 - i1 == 1
                    and j2 - j1 == 1
                ):
                    # logic is from get_merge_wdiff2(chunk_index)
                    """Return content for wdiff by line (2 files)"""
                    (
                        tag,
                        i1,
                        i2,
                        j1,
                        j2,
                        k1,
                        k2,
                        action,
                        merge_buffer,
                    ) = self.chunk_list[
                        chunk_index
                    ]  # chunk_list item tuple (9 param)
                    logger.debug(
                        "chunk[{}]: tag={} === a[{}:{}]/b[{}:{}]/_[{}:{}] action='{}' len(merge_buffer)={}".format(
                            chunk_index,
                            tag,
                            i1,
                            i2,
                            j1,
                            j2,
                            k1,
                            k2,
                            action,
                            len(merge_buffer),
                        ),
                    )
                    line_a = self.list_a[i1]
                    line_b = self.list_b[j1]
                    if self.isjunk:
                        isjunk = None
                    else:
                        isjunk = self.whitespace_is_junk
                    matcher_internal = SequenceMatcher(isjunk, line_a, line_b, False)
                    chunk_list_internal = matcher_internal.get_opcodes()
                    content = list()
                    for tag, i1, i2, j1, j2 in chunk_list_internal:
                        if tag == "equal":
                            # wdiff-clean
                            content.append(
                                (
                                    line_a[i1:i2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_ab", focus),
                                )
                            )
                        else:  # other tags (mark up with word separator)
                            content.append(
                                (
                                    self.ws0,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                            content.append(
                                (
                                    line_a[i1:i2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_a", focus),
                                )
                            )
                            content.append(
                                (
                                    self.ws1,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                            content.append(
                                (
                                    line_b[j1:j2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_b2", focus),
                                )
                            )
                            content.append(
                                (
                                    self.ws3,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                    del chunk_list_internal
                    del matcher_internal
                    # content = basically list of attribute added text of get_merge_wdiff2(chunk_index)
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        content,
                        "f",
                        clrtoeol=True,
                    )
                elif (
                    action == "f"
                    and self.diff_mode == 3
                    and i2 - i1 == 1
                    and j2 - j1 == 1
                    and k2 - k1 == 1
                ):
                    # always non-clean merge (clean merge is in e for merge_buffer)
                    # logic is from get_merge_wdiff3(chunk_index)
                    """Return content for wdiff by line (3 files)"""
                    (
                        tag,
                        i1,
                        i2,
                        j1,
                        j2,
                        k1,
                        k2,
                        action,
                        merge_buffer,
                    ) = self.chunk_list[
                        chunk_index
                    ]  # chunk_list item tuple (9 param)
                    if i2 - i1 != 1 or j2 - j1 != 1 or k2 - k1 != 1:
                        logger.error(
                            "chunk[{}]: tag={} === a[{}:{}]/b[{}:{}]/c[{}:{}] not for wdiff3".format(
                                chunk_index, tag, i1, i2, j1, j2, k1, k2
                            ),
                        )
                        sys.exit(2)
                    logger.debug(
                        "chunk[{}]: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] action='{}' wdiff3-merge-try".format(
                            chunk_index, tag, i1, i2, j1, j2, k1, k2, action
                        ),
                    )
                    line_a = self.list_a[i1]
                    line_b = self.list_b[j1]
                    line_c = self.list_c[k1]
                    if self.isjunk:
                        isjunk = None
                    else:
                        isjunk = self.whitespace_is_junk
                    # wdiff uses SequenceMatcher
                    use_SequenceMatcher = 0
                    matcher_internal = SequenceMatcher3(
                        line_a, line_b, line_c, use_SequenceMatcher, isjunk, True
                    )
                    chunk_list_internal = matcher_internal.get_opcodes()
                    # logger.debug("wdiff3: \nwchunk_list_internal >>>>> {}".format(wchunk_list_internal))
                    content = list()
                    clean_merge = True
                    for tag, i1, i2, j1, j2, k1, k2 in chunk_list_internal:
                        if tag == "A":
                            content.append(
                                (
                                    line_a[i1:i2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_a", focus),
                                )
                            )
                        elif tag == "C":
                            content.append(
                                (
                                    line_c[k1:k2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_c", focus),
                                )
                            )
                        elif tag == "E":
                            content.append(
                                (
                                    line_c[k1:k2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_abc", focus),
                                )
                            )
                        elif tag == "e":
                            content.append(
                                (
                                    line_c[k1:k2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_ac", focus),
                                )
                            )
                        else:  # tag == "N" (mark up with word separator)
                            content.append(
                                (
                                    self.ws0,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                            content.append(
                                (
                                    line_a[i1:i2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_a", focus),
                                )
                            )
                            content.append(
                                (
                                    self.ws1,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                            content.append(
                                (
                                    line_b[j1:j2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_b3", focus),
                                )
                            )
                            content.append(
                                (
                                    self.ws2,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                            content.append(
                                (
                                    line_c[k1:k2],
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_c", focus),
                                )
                            )
                            content.append(
                                (
                                    self.ws3,
                                    corner_virt_col,
                                    corner_virt_col + stdscr_col_max,
                                    self.get_attr("color_wdiff_marker", focus),
                                )
                            )
                    del matcher_internal
                    del chunk_list_internal
                    logger.debug(
                        "chunk[{}]: clean_merge={} === tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] action='{}'".format(
                            chunk_index,
                            clean_merge,
                            tag,
                            i1,
                            i2,
                            j1,
                            j2,
                            k1,
                            k2,
                            action,
                        ),
                    )
                    # (clean_merge, content) = self.get_merge_wdiff3(chunk_index)
                    self.display_imediff_content(
                        self.stdscr,
                        row_index,
                        content,
                        "f",
                        clrtoeol=True,
                    )
                else:
                    pass  # but error
                    # TODO: add log error
            else:
                display_content(
                    self.stdscr,
                    row_index,
                    0,
                    [
                        ("  ", 0, 2, self.get_attr("color_white", False)),
                        (
                            "...[EOF]...",
                            0,
                            stdscr_col_max,
                            self.get_attr("color_eof", False),
                        ),
                    ],
                    clrtoeol=True,
                )
            #
        #
        # STATISTICS
        # * s_chunk_index
        # * len(chunk_list)
        # * self.focused_usr_chunk_index
        n_merge_E = 0  # = =
        n_merge_e = 0  #   #
        n_merge_n = 0  #   G
        n_merge_A = 0  #   A
        n_merge_C = 0  #   C
        n_merge_N = 0  # includes F for diff2
        n_manual_a = 0
        n_manual_b = 0
        n_manual_c = 0
        n_manual_e = 0
        n_unresolved = 0
        if self.focused_usr_chunk_index is not None:
            focused_chunk_index = self.get_chunk_index_from_usr_chunk_list(
                self.focused_usr_chunk_index
            )
            s_focused_chunk_index = s_number(focused_chunk_index)
            s_focused_usr_chunk_index = s_number(self.focused_usr_chunk_index)
            if focused_chunk_index is None:
                s_virt_row = "*"
            else:
                s_virt_row = s_number(self.chunk_to_virt[focused_chunk_index])
            for chunk_index in range(len(self.chunk_list)):
                tag = self.get_tag(chunk_index)
                action = self.get_action(chunk_index)
                if tag == "E":
                    n_merge_E += 1
                elif tag == "e":
                    n_merge_e += 1
                elif tag == "n":
                    n_merge_n += 1
                elif tag == "A":
                    n_merge_A += 1
                elif tag == "C":
                    n_merge_C += 1
                else:  # N or F
                    n_merge_N += 1
                    if action == "a":
                        n_manual_a += 1
                    elif action == "b":
                        n_manual_b += 1
                    elif action == "c":
                        n_manual_c += 1
                    elif action == "e":
                        n_manual_e += 1
                    else:
                        n_unresolved += 1
            s_merge_E = s_number(n_merge_E)
            s_merge_e = s_number(n_merge_e)
            s_merge_n = s_number(n_merge_n)
            s_merge_A = s_number(n_merge_A)
            s_merge_C = s_number(n_merge_C)
            s_merge_N = s_number(n_merge_N)
            s_manual_a = s_number(n_manual_a)
            s_manual_b = s_number(n_manual_b)
            s_manual_c = s_number(n_manual_c)
            s_manual_e = s_number(n_manual_e)
            s_unresolved = s_number(n_unresolved)
        else:
            s_focused_chunk_index = "*"
            s_focused_usr_chunk_index = "*"
            for chunk_index in range(len(self.chunk_list)):
                tag = self.get_tag(chunk_index)
                action = self.get_action(chunk_index)
                if tag == "E":
                    n_merge_E += 1
                elif tag == "e":
                    n_merge_e += 1
                elif tag == "n":
                    n_merge_n += 1
                elif tag == "A":
                    n_merge_A += 1
                elif tag == "C":
                    n_merge_C += 1
                else:  # N or F
                    n_merge_N += 1
            s_merge_E = s_number(n_merge_E)
            s_merge_e = s_number(n_merge_e)
            s_merge_n = s_number(n_merge_n)
            s_merge_A = s_number(n_merge_A)
            s_merge_C = s_number(n_merge_C)
            s_merge_N = s_number(n_merge_N)
            s_virt_row = "*"
            s_manual_a = "*"
            s_manual_b = "*"
            s_manual_c = "*"
            s_manual_e = "*"
            s_unresolved = "*"

        # len(self.usr_chunk_list)
        #
        if self.diff_mode == 2:
            status_line = "row[{}/{}] chunk[{}/{}] usr_chunk[{}/{}] / =:{} / N:{}=(a:{},b:{},e:{},u:{}) / @[{}:{}]".format(
                s_virt_row,
                len(self.virt_to_chunk),
                s_focused_chunk_index,
                len(self.chunk_list),
                s_focused_usr_chunk_index,
                len(self.usr_chunk_list),
                s_merge_E,
                s_merge_N,
                s_manual_a,
                s_manual_b,
                s_manual_e,
                s_unresolved,
                corner_virt_row,
                corner_virt_col,
            )
        else:
            status_line = "row[{}/{}] chunk[{}/{}] usr_chunk[{}/{}] / =:{},#:{},G:{},A:{},C:{} / N:{}=(a:{},b:{},c:{},e:{},u:{}) / @[{}:{}]".format(
                s_virt_row,
                len(self.virt_to_chunk),
                s_focused_chunk_index,
                len(self.chunk_list),
                s_focused_usr_chunk_index,
                len(self.usr_chunk_list),
                s_merge_E,
                s_merge_e,
                s_merge_n,
                s_merge_A,
                s_merge_C,
                s_merge_N,
                s_manual_a,
                s_manual_b,
                s_manual_c,
                s_manual_e,
                s_unresolved,
                corner_virt_row,
                corner_virt_col,
            )
        display_content(
            self.stdscr,
            stdscr_row_max - 1,
            0,
            [(status_line, 0, None, self.get_attr("color_white_reverse", False))],
            clrtoeol=True,
        )
        self.stdscr.refresh()

    ####################################################################
    def display_imediff_content(self, win, row, content, data_source, clrtoeol=True):
        # data_source = "a", "b", "c", "=", "#", "f", "d??", "G", "e", "A", "C"
        col = 0
        if data_source[:1] == "d":
            data_source_id = " "
            color = "color_status_focus"
        elif data_source in "GeAC=#":
            data_source_id = data_source
            color = "color_status"
        else:
            data_source_id = data_source
            color = "color_status_focus"
        if True:  # self.data_source_column:
            display_content(
                win,
                row,
                col,
                [
                    (
                        data_source_id,
                        0,
                        1,
                        self.get_attr(color, False),
                    ),
                    (" ", 0, 1, self.get_attr("color_white", False)),
                ],
                False,
            )
            col = None
        for line, i_b, i_e, attrib in content:
            display_content(
                win,
                row,
                col,
                [(line, i_b, i_e, attrib)],
                clrtoeol,
            )
            col = None

    ####################################################################
    # Internally used utility methods (class data set-access)
    ####################################################################

    def set_updated_merge_buffer(self, chunk_index):  # override
        logger.debug(
            "chunk[{}]: exit the curses UI and to invoke editor session".format(
                chunk_index
            )
        )
        self.stdscr.keypad(False)  # keys not-processed by curses
        curses.savetty()
        curses.echo()
        curses.nocbreak()
        curses.endwin()
        super().set_updated_merge_buffer(chunk_index)
        curses.cbreak()
        curses.noecho()
        curses.resetty()
        self.stdscr.keypad(True)  # keys processed by curses (again)
        self.stdscr.clear()
        self.stdscr.refresh()
        logger.debug(
            "chunk[{}]: finish editor session and return to the curses UI".format(
                chunk_index
            )
        )
        return

    def display_popup_win(
        self,
        msg,
        keyname_list=["q"],
        color="color_warn",
        msg_row_max=0,
        msg_col_max=0,
        margin_row=1,
        margin_col=2,
    ):
        msg_content = list()
        for line in msg.split("\n"):  # no trailing \n
            msg_content.append((line, self.get_attr(color, False)))

        keyname = self.display_content_win(
            msg_content, keyname_list, msg_row_max, msg_col_max, margin_row, margin_col
        )
        return keyname

    def display_content_win(
        self,
        msg_content,
        keyname_list=["q"],
        msg_row_max=0,
        msg_col_max=0,
        margin_row=1,
        margin_col=2,
    ):
        # A pop-up centered window with specified window size and border box
        # drawing.  The text are drawn in the box and has 1 space before and after
        # for readability.
        #
        # get msg display size for virt_msg
        virt_row_max = len(msg_content)
        virt_col_max = msg_col_max
        for line, _ in msg_content:
            virt_col_max = max(virt_col_max, len(line))
        # set up centered box_win
        stdscr_row_max, stdscr_col_max = self.stdscr.getmaxyx()
        # usable msg display span
        span_col = stdscr_col_max - 2 * margin_col
        span_row = stdscr_row_max - 2 * margin_row
        logger.debug(
            "stdscr={}:{} virt={}:{} span={}:{}".format(
                stdscr_row_max,
                stdscr_col_max,
                virt_row_max,
                virt_col_max,
                span_row,
                span_col,
            )
        )
        if msg_row_max == 0 and (virt_row_max + 2 * margin_row) <= stdscr_row_max:
            msg_row_max = virt_row_max
        elif msg_row_max == 0 or msg_row_max >= span_row:
            # force reasonable msg_row_max
            msg_row_max = stdscr_row_max - 2 * margin_row
        if msg_col_max == 0 and (virt_col_max + 2 * margin_col) <= stdscr_col_max:
            msg_col_max = virt_col_max
        elif msg_col_max == 0 or msg_col_max >= span_col:
            # force reasonable msg_col_max
            msg_col_max = stdscr_col_max - 2 * margin_col
        box_margin_row = (stdscr_row_max - (msg_row_max + 2 * margin_row)) // 2
        box_margin_col = (stdscr_col_max - (msg_col_max + 2 * margin_col)) // 2
        self.box_win = self.stdscr.derwin(
            msg_row_max + 2 * margin_row,
            msg_col_max + 2 * margin_col,
            box_margin_row,
            box_margin_col,
        )
        self.box_win.clear()
        self.box_win.border()
        self.box_win.refresh()
        # set up centered msg_win
        self.msg_win = self.box_win.derwin(
            msg_row_max, msg_col_max, margin_row, margin_col
        )
        corner_popup_row = 0
        corner_popup_col = 0
        while True:
            msg_row_max, msg_col_max = self.msg_win.getmaxyx()
            corner_popup_row = max(0, corner_popup_row)
            corner_popup_col = max(0, corner_popup_col)
            self.msg_win.move(0, 0)
            for row_index in range(msg_row_max):
                virt_row_index = row_index + corner_popup_row
                if virt_row_index < len(msg_content):
                    line, attr = msg_content[virt_row_index]
                    display_content(
                        self.msg_win,
                        row_index,
                        0,
                        [(line[corner_popup_col:], 0, None, attr)],
                        clrtoeol=True,
                    )
                else:
                    display_content(
                        self.msg_win,
                        row_index,
                        0,
                        [
                            (
                                "...[EOF]...",
                                0,
                                msg_col_max,
                                self.get_attr("color_eof", False),
                            )
                        ],
                        clrtoeol=True,
                    )
            self.msg_win.refresh()
            c = self.stdscr.getch()  # c : integer (stdscr! here)
            keyname = get_keyname(c)
            if keyname in keyname_list:
                break
            # Moves in document
            elif keyname in ["0", "HOME"]:
                corner_popup_row = 0
                corner_popup_col = 0
            elif keyname in ["9", "END"]:
                corner_popup_row = max(0, virt_row_max - msg_row_max)
                corner_popup_col = 0
            elif keyname in ["j", "DOWN"]:
                corner_popup_row += 1
            elif keyname in ["k", "UP"]:
                corner_popup_row -= 1
            elif keyname in ["l", "RIGHT"]:
                corner_popup_col += 8
            elif keyname in ["h", "LEFT"]:
                corner_popup_col -= 8
            elif keyname in ["n", "PAGEDOWN", "SPACE", "ENTER"]:
                corner_popup_row += 20
            elif keyname in ["p", "PAGEUP", "BACKSPACE"]:
                corner_popup_row -= 20
            else:
                pass
        return keyname

    ####################################################################
    # Internally used utility methods (user feedback)
    ####################################################################
    def report(self, message):  # override for TUI (simple pop-up)
        self.display_popup_win(message, ["SPACE", "ESCAPE", "q", "Q"])



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

import curses
import sys
import io
import logging

logger = logging.getLogger(__name__)

help_msg = "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\n" * 30 + "ZZZZ"

text_example = """\
What IS lOREM Ipsum? lOREM Ipsum IS simply dummy text of the printing and
typesetting industry. lOREM Ipsum has been the industry's standard dummy text
ever since the 1500s, when an unknown printer took a galley of type and
scrambled it to make a type specimen book. It has survived not only five
centuries, but also the leap into electronic typesetting, remaining essentially
unchanged. It was popularISed in the 1960s with the release of Letraset sheets
containing lOREM Ipsum passages, and more recently with desktop publIShing
software like Aldus PageMaker including versions of lOREM Ipsum.

EOF_LINE
"""


class TUI:
    """Curses class to handle text"""

    def __init__(self, text):
        self.virt_lines = text.splitlines(keepends=True)
        return

    ####################################################################
    # Externally used main method and effective main method
    ####################################################################
    # Design concept:
    #     * Use full terminal screen
    #     * Access them directly (no window nor pad)
    #     * String operation by pointers
    #
    # See https://docs.python.org/3/library/curses.html
    #     https://docs.python.org/3/howto/curses.html
    #
    #     curses terminal screen absolute coordinate as (y,x):
    #     +----------------------------------------------------------+
    #     | (0,0) top-left corner                    (term_col_max-1)|
    #     |  +--> x-axis                                             |
    #     |  |                                                       |
    #     |  |                 TEXT DATA DISPLAY                     |
    #     |  V y-axis                                                |
    #     |                                                          |
    #     |                                                          |
    #     | (term_row_max-2,0)        (term_row_max-2,term_col_max-1)|
    #     +----------------------------------------------------------+
    #     | (term_row_max-1,0) STATUS (term_row_max-1,term_col_max-1)|
    #     +----------------------------------------------------------+
    #
    # Here:
    #     * term_row_max, term_col_max = stdscr.getmaxyx()
    #
    # Internal data:
    #   * TUI
    #     * virt_to_row[virt_row] = (chunk_index, chunk_subindex)
    # Terminal size: 80 col x 24 row required
    #
    ####################################################################

    def main(self):  # for curses TUI (wrapper)
        """CURSES TUI program"""
        logger.debug("tui-main start")
        curses.wrapper(self.tui_loop)
        logger.debug("tui-main end")
        return

    def tui_loop(self, stdscr):  # for curses TUI (core)
        """wrapped CURSES TUI loop"""
        self.stdscr = stdscr
        curses.curs_set(0)  # cursor off
        # create text_win and status_win
        self.init_curses()
        # initial display parameters
        # flag_update_text_win_corner = True
        self.text_win_corner_virt_row = 0
        while True:
            # enforce sanity
            self.text_win_corner_virt_row = max(0, self.text_win_corner_virt_row)
            self.display_text_win()
            c = self.stdscr.getch()  # return integer
            ch = chr(c)
            if ch == "q":
                logger.debug("quit without write to a file: imediff.out")
                break
            # elif ch == "h" or c == curses.KEY_HELP:
            # Show help screen
            # self.display_popup_win(self.get_helptext())
            # Moves in document
            elif c == curses.KEY_HOME or ch == "0":
                self.text_win_corner_virt_row = 0
            elif c == curses.KEY_SF or c == curses.KEY_DOWN or ch == "j":
                self.text_win_corner_virt_row += 1
            elif c == curses.KEY_SR or c == curses.KEY_UP or ch == "k":
                self.text_win_corner_virt_row -= 1
            elif c == curses.KEY_NPAGE or ch == ";":
                self.text_win_corner_virt_row += 20
            elif c == curses.KEY_PPAGE or ch == "'":
                self.text_win_corner_virt_row -= 20
            # Terminal resize signal
            else:
                pass
            logger.debug("gui-command-loop")
        return

    ####################################################################
    def init_curses(self):
        # curses.start_color() # wrapper takes care
        logger.debug(
            "COLORS={}, COLOR_PAIRS={}".format(curses.COLORS, curses.COLOR_PAIRS)
        )
        stdscr_row_max, stdscr_col_max = self.stdscr.getmaxyx()
        logger.debug(
            "stdscr_row_max={}, stdscr_col_max={}".format(
                stdscr_row_max, stdscr_col_max
            )
        )
        if stdscr_row_max < 20 or stdscr_col_max < 60:
            logger.error(
                "E: terminal size too small: row={}:col={}".format(
                    stdscr_row_max, stdscr_col_max
                )
            )
            sys.exit(2)
        self.stdscr.clear()
        self.stdscr.refresh()
        self.text_win = curses.newwin(stdscr_row_max - 1, stdscr_col_max)
        self.status_win = curses.newwin(1, stdscr_col_max, stdscr_row_max - 1, 0)
        # set up color
        # curses.init_pair(0, curses.COLOR_WHITE, curses.COLOR_BLACK)
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(4, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
        curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_BLACK)
        curses.init_pair(8, curses.COLOR_WHITE, curses.COLOR_BLACK)
        curses.init_pair(9, curses.COLOR_WHITE, curses.COLOR_WHITE)
        curses.init_pair(10, curses.COLOR_RED, curses.COLOR_WHITE)
        curses.init_pair(11, curses.COLOR_GREEN, curses.COLOR_WHITE)
        curses.init_pair(12, curses.COLOR_YELLOW, curses.COLOR_WHITE)
        curses.init_pair(13, curses.COLOR_BLUE, curses.COLOR_WHITE)
        curses.init_pair(14, curses.COLOR_MAGENTA, curses.COLOR_WHITE)
        curses.init_pair(15, curses.COLOR_CYAN, curses.COLOR_WHITE)
        curses.init_pair(16, curses.COLOR_BLACK, curses.COLOR_WHITE)
        self.content = [
            (
                curses.color_pair(0),
                "curses.init_pair(0_ -- default NORMAL >>>>>>>>>>>>>>>",
            ),
            (
                curses.color_pair(1),
                "curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(2),
                "curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(3),
                "curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(4),
                "curses.init_pair(4, curses.COLOR_BLUE, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(5),
                "curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(6),
                "curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(7),
                "curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(8),
                "curses.init_pair(8, curses.COLOR_WHITE, curses.COLOR_BLACK) PREV=B_on_B, NEXT=W-on_W",
            ),
            (
                curses.color_pair(9),
                "curses.init_pair(9, curses.COLOR_WHITE, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(10),
                "curses.init_pair(10, curses.COLOR_RED, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(11),
                "curses.init_pair(11, curses.COLOR_GREEN, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(12),
                "curses.init_pair(12, curses.COLOR_YELLOW, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(13),
                "curses.init_pair(13, curses.COLOR_BLUE, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(14),
                "curses.init_pair(14, curses.COLOR_MAGENTA, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(15),
                "curses.init_pair(15, curses.COLOR_CYAN, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(16),
                "curses.init_pair(16, curses.COLOR_BLACK, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(0) | curses.A_REVERSE,
                "curses.init_pair(0_ -- default REVERSE >>>>>>>>>>>>>>>",
            ),
            (
                curses.color_pair(1) | curses.A_REVERSE,
                "curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(2) | curses.A_REVERSE,
                "curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(3) | curses.A_REVERSE,
                "curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(4) | curses.A_REVERSE,
                "curses.init_pair(4, curses.COLOR_BLUE, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(5) | curses.A_REVERSE,
                "curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(6) | curses.A_REVERSE,
                "curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(7) | curses.A_REVERSE,
                "curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(8) | curses.A_REVERSE,
                "curses.init_pair(8, curses.COLOR_WHITE, curses.COLOR_BLACK) PREV=B_on_B, NEXT=W-on_W",
            ),
            (
                curses.color_pair(9) | curses.A_REVERSE,
                "curses.init_pair(9, curses.COLOR_WHITE, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(10) | curses.A_REVERSE,
                "curses.init_pair(10, curses.COLOR_RED, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(11) | curses.A_REVERSE,
                "curses.init_pair(11, curses.COLOR_GREEN, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(12) | curses.A_REVERSE,
                "curses.init_pair(12, curses.COLOR_YELLOW, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(13) | curses.A_REVERSE,
                "curses.init_pair(13, curses.COLOR_BLUE, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(14) | curses.A_REVERSE,
                "curses.init_pair(14, curses.COLOR_MAGENTA, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(15) | curses.A_REVERSE,
                "curses.init_pair(15, curses.COLOR_CYAN, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(16) | curses.A_REVERSE,
                "curses.init_pair(16, curses.COLOR_BLACK, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(0) | curses.A_STANDOUT,
                "curses.init_pair(0_ -- default STANDOUT >>>>>>>>>>>>>>>",
            ),
            (
                curses.color_pair(1) | curses.A_STANDOUT,
                "curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(2) | curses.A_STANDOUT,
                "curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(3) | curses.A_STANDOUT,
                "curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(4) | curses.A_STANDOUT,
                "curses.init_pair(4, curses.COLOR_BLUE, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(5) | curses.A_STANDOUT,
                "curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(6) | curses.A_STANDOUT,
                "curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(7) | curses.A_STANDOUT,
                "curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_BLACK)",
            ),
            (
                curses.color_pair(8) | curses.A_STANDOUT,
                "curses.init_pair(8, curses.COLOR_WHITE, curses.COLOR_BLACK) PREV=B_on_B, NEXT=W-on_W",
            ),
            (
                curses.color_pair(9) | curses.A_STANDOUT,
                "curses.init_pair(9, curses.COLOR_WHITE, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(10) | curses.A_STANDOUT,
                "curses.init_pair(10, curses.COLOR_RED, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(11) | curses.A_STANDOUT,
                "curses.init_pair(11, curses.COLOR_GREEN, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(12) | curses.A_STANDOUT,
                "curses.init_pair(12, curses.COLOR_YELLOW, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(13) | curses.A_STANDOUT,
                "curses.init_pair(13, curses.COLOR_BLUE, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(14) | curses.A_STANDOUT,
                "curses.init_pair(14, curses.COLOR_MAGENTA, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(15) | curses.A_STANDOUT,
                "curses.init_pair(15, curses.COLOR_CYAN, curses.COLOR_WHITE)",
            ),
            (
                curses.color_pair(16) | curses.A_STANDOUT,
                "curses.init_pair(16, curses.COLOR_BLACK, curses.COLOR_WHITE)",
            ),
            (curses.color_pair(0), "curses.init_pair(0_ -- default =================="),
        ]

    ####################################################################
    def display_text_win(self):
        text_win_row_max, text_win_col_max = self.text_win.getmaxyx()
        for row_index in range(text_win_row_max):
            virt_row_index = row_index + self.text_win_corner_virt_row
            if virt_row_index < len(self.content):
                # valid row
                self.text_win.addstr(
                    row_index,
                    0,
                    self.content[virt_row_index][1][: text_win_col_max - 2],
                    self.content[virt_row_index][0],
                )
                self.text_win.clrtoeol()
            else:
                # invalid row
                self.text_win.addstr(
                    row_index,
                    0,
                    "...[EOF]...",
                    curses.color_pair(4) | curses.A_DIM,
                )
                self.text_win.clrtoeol()
        self.text_win.refresh()


if __name__ == "__main__":
    logging.basicConfig(
        format="%(levelname)s: %(filename)s: %(funcName)s: %(message)s",
        filename="example.log",
        level=logging.DEBUG,
    )
    logger.debug("__main__ start")
    tui = TUI(text_example * 3)
    tui.main()
    # finish
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

import argparse

# This allows different reporting style for TUI/CLI
pa = argparse.ArgumentParser(
    description="SIMPLE example",
    epilog="See README.md for more",
)
group = pa.add_mutually_exclusive_group()
group.add_argument(
    "-a", action="store_true", help="Start with all chunks to use file_a"
)
group.add_argument(
    "-b", action="store_true", help="Start with all chunks to use file_b"
)
group.add_argument(
    "-c",
    action="store_true",
    help="Start with all chunks to use file_c (only for diff3)",
)
pa.add_argument(
    "--rule-filter",
    "-r",
    action="store",
    default=2,
    help="Fuzzy match line filtering rule (0,1,2,3,10,11,12,13)",
)
args = pa.parse_args()

print("a={}, b={}, c={}".format(args.a, args.b, args.c))
print("rule-filter={}".format(args.rule_filter))



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai:
"""
Script to test diff2 and diff3 logic

The diff3 logic is build on top of diff2 logic.

Current diff2 uses Python standard library difflib which uses a variant of
longest contiguous matching subsequence algorithm by Ratcliff and Obershelp
developed in the late 1980's.  If I update this imediff program to use more
modern algorithm, this test may yield slightly different result in some
corner cases.
"""

import os
import sys
import difflib
import imediff.diff3lib


def diff23(a, b, c, name=""):
    print("# Test type: " + name)
    print("  a='%s' -> %i" % (a, len(a)))
    print("  b='%s' -> %i" % (b, len(b)))
    print("  c='%s' -> %i" % (c, len(c)))
    sa = difflib.SequenceMatcher(None, a, b)
    print("$ diff2 A B")
    for tag, i1, i2, j1, j2 in sa.get_opcodes():
        print(
            (
                "  %s     a[%d:%d] (%s) / b[%d:%d] (%s)"
                % (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2])
            )
        )
    print("$ diff2 C B")
    sc = difflib.SequenceMatcher(None, c, b)
    for tag, i1, i2, j1, j2 in sc.get_opcodes():
        print(
            (
                "  %s     c[%d:%d] (%s) / b[%d:%d] (%s)"
                % (tag, i1, i2, c[i1:i2], j1, j2, b[j1:j2])
            )
        )
    s = imediff.diff3lib.SequenceMatcher3(a, b, c, 0, None, True)
    print("$ diff3 A B C")
    for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
        print(
            (
                "  %s     a[%d:%d] (%s) / b[%d:%d] (%s) / c[%d:%d] (%s)"
                % (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2], k1, k2, c[k1:k2])
            )
        )
    print("===========================================================")
    return


print("\nI: _diff23 start >>>", file=sys.stderr)
cwd_dir = os.getcwd()
print("I: cwd_dir     = '{}'".format(cwd_dir), file=sys.stderr)
print("I: test_file   = '{}' (active)".format(__file__), file=sys.stderr)
test_dir = os.path.dirname(os.path.abspath(__file__))
print("I: test_dir    = '{}' (active)".format(test_dir), file=sys.stderr)
if "PYTHONPATH" in os.environ:
    print("I: PYTHONPATH  = '{}'".format(os.environ["PYTHONPATH"]), file=sys.stderr)
else:
    print("I: PYTHONPATH  = <undefined>", file=sys.stderr)
print("I: _diff23 end   <<<", file=sys.stderr)

diff23("12345", "12345", "12345", "same")

diff23("a12345z", "1245", "1245", "add a side")

diff23("a12345z", "1245", "a12345z", "add same")

diff23("1245", "1245", "a12345z", "add c side")

diff23("24", "12345", "12345", "delete a side")

diff23("24", "12345", "24", "delete same")

diff23("12345", "12345", "24", "delete c side")

diff23("a2b4c", "12345", "x2y4z", "conflict both ends")

diff23("1a2b4c", "12345", "x2y45z", "conflict skewed")



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
Copyright (C) 2024 Osamu Aoki <osamu@debian.org>
"""

import sys
import curses
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(filename="example.log", level=logging.DEBUG)
exception = None
# See https://docs.python.org/3/library/curses.html
#     https://docs.python.org/3/howto/curses.html
# logger.debug('This message should go to the log file')
# logger.info('So should this')
# logger.warning('And this, too')
# logger.error('And non-ASCII stuff, too, like Øresund and Malmö')


def test_text(char, height, width):
    test_text = ""
    for col in range(0, width):
        test_text += "{:1}".format(col % 10)
    test_text += "\n\n\n"
    for col in range(0, width):
        test_text += "{:1}".format((col // 10) % 10)
    test_text += "\n\n\n"
    for row in range(6, (height // 3) - 6):
        test_text += (char[:1] * 32 + "  {:4}  ".format(row) + char[:1] * 40) * (
            width // 80
        ) + "\n\n\n"
    for row in range(height - (height // 3) * 3, height):
        test_text += "{:1}\n".format(char[:1] * width)
    return test_text


####################################################################
# Externally used main method and effective main method gui_loop
####################################################################
def main():  # for curses wrapper
    global exception
    debug("start CURSES TEST")
    curses.wrapper(gui_main)
    if exception is not None:
        print("E: {} {}".format(type(exception).__name__, exception))
    debug("end CURSES TEST")
    return


def gui_main(stdscr):  # for curses TUI (core)
    global exception
    # Clear screen
    try:
        curses.curs_set(0)  # cursor off
    except Exception as e:
        exception = e
    # https://stackoverflow.com/questions/3170406/python-curses-newwin-not-working
    stdscr.clear()
    # curses.curs_set(1)  # cursor on
    curses.curs_set(0)  # cursor off
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(5, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)
    curses.init_pair(7, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE)
    # This raises ZeroDivisionError when i == 10.
    term_row_max, term_col_max = stdscr.getmaxyx()
    stdscr.insstr(term_row_max - 1, term_col_max - 1, "@")
    stdscr.refresh()
    stdscr.getkey()
    textarea = stdscr.derwin(term_row_max - 1, term_col_max, 0, 0)
    textarea_row_max, textarea_col_max = textarea.getmaxyx()
    textarea.insstr(textarea_row_max - 1, textarea_col_max - 1, "*")
    textarea.refresh()
    textarea.getkey()
    statusarea = stdscr.derwin(1, term_col_max, term_row_max - 1, 0)
    statusarea_row_max, statusarea_col_max = statusarea.getmaxyx()
    statusarea.insstr(statusarea_row_max - 1, statusarea_col_max - 1, "%")
    statusarea.refresh()
    statusarea.getkey()
    statusarea = stdscr.derwin(1, term_col_max, term_row_max - 1, 0)
    statusarea_row_max, statusarea_col_max = statusarea.getmaxyx()
    try:
        statusarea.addstr(statusarea_row_max - 1, statusarea_col_max - 1, "$")
    except curses.error as _:
        pass
    statusarea.refresh()
    # getkey is noy window specific
    # statusarea.getkey()
    # textarea.getkey()
    stdscr.getkey()
    stdscr.clear()
    stdscr.border()
    stdscr.refresh()
    stdscr.getkey()
    for row in range(0, textarea_row_max):
        color = row % 6
        textarea.insstr(row, 0, "&" * (textarea_col_max), color)
    textarea.addstr(textarea_row_max - 1, 10, "!" * 10)
    textarea.clrtoeol()
    textarea.refresh()
    stdscr.getkey()
    offset_col = 10
    stdscr.addstr(term_row_max - 1, offset_col, "Z" * (term_col_max - 1 - offset_col))
    stdscr.insch("E")
    stdscr.refresh()
    stdscr.getkey()

    #####################################################################################


def get_center_margin(span_out, span_in, margin_min):
    span_in_real = min(span_out - 2 * margin_min, span_in)
    margin = (span_out - span_in_real) // 2
    return margin


def get_sanitized(x, x_min, x_max):
    if x_min > x_max:
        print("ERROR min={}, max={}".format(x_min, x_max))
        sys.exit(1)
    elif x < x_min:
        x = x_min
    elif x >= x_max:
        x = x_max
    else:
        pass  # sanity OK
    return x


####################################################################
# Internally used utility methods (popup display via curses)
####################################################################
def tui_popup_text(stdscr, text):
    curses.curs_set(0)
    term_row_max, term_col_max = stdscr.getmaxyx()
    stdscr.border()
    for line_index, line_text in enumerate(text.split("\n")):  # no \n
        if line_index < term_row_max - 5:
            stdscr.addstr(
                line_index + 1, 2, line_text[: term_col_max - 5], curses.A_BOLD
            )
    stdscr.refresh()
    stdscr.getch()
    return


# debug message
def debug(msg):
    logger.debug(msg)
    return


if __name__ == "__main__":
    # run program
    main()
    # finish
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
Copyright (C) 2024 Osamu Aoki <osamu@debian.org>
"""

import sys
import curses
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(filename="example.log", level=logging.DEBUG)
exception = None
# See https://docs.python.org/3/library/curses.html
#     https://docs.python.org/3/howto/curses.html
# logger.debug('This message should go to the log file')
# logger.info('So should this')
# logger.warning('And this, too')
# logger.error('And non-ASCII stuff, too, like Øresund and Malmö')


def test_text(char, height, width):
    test_text = ""
    for col in range(0, width):
        test_text += "{:1}".format(col % 10)
    test_text += "\n\n\n"
    for col in range(0, width):
        test_text += "{:1}".format((col // 10) % 10)
    test_text += "\n\n\n"
    for row in range(6, (height // 3) - 6):
        test_text += (char[:1] * 32 + "  {:4}  ".format(row) + char[:1] * 40) * (
            width // 80
        ) + "\n\n\n"
    for row in range(height - (height // 3) * 3, height):
        test_text += "{:1}\n".format(char[:1] * width)
    return test_text


####################################################################
# Externally used main method and effective main method gui_loop
####################################################################
def main():  # for curses wrapper
    global exception
    logger.debug("start CURSES TEST")
    curses.wrapper(gui_main)
    if exception is not None:
        print("E: {} {}".format(type(exception).__name__, exception))
    logger.debug("end CURSES TEST")
    return


def gui_main(stdscr):  # for curses TUI (core)
    global exception
    y0, x0 = stdscr.getyx()
    # curses.curs_set(1)  # cursor on
    curses.curs_set(0)  # cursor off
    stdscr.clear()
    # curses.start_color()
    # curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)  # mode=e
    curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)  # tag=E:DIM/e:NORMAL
    curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)  # marker (diff/nill)
    curses.init_pair(4, curses.COLOR_YELLOW, curses.COLOR_BLACK)  # mode=f
    curses.init_pair(5, curses.COLOR_GREEN, curses.COLOR_BLACK)  # mode=a
    curses.init_pair(6, curses.COLOR_CYAN, curses.COLOR_BLACK)  # mode=b
    curses.init_pair(7, curses.COLOR_MAGENTA, curses.COLOR_BLACK)  # mode=c
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE)  # NO
    margin = 2
    term_row_max, term_col_max = stdscr.getmaxyx()
    win_row_max = term_row_max - margin * 2 - 1
    win_col_max = term_col_max - margin * 2
    test_lines = test_text("#", term_row_max, term_col_max)
    for i, test_line in enumerate(test_lines.split("\n")):
        if i + margin < term_row_max - (margin + 1):
            stdscr.addstr(i + margin, margin, test_line[: term_col_max - margin * 2])
    stdscr.addstr(
        term_row_max - 1,
        0,
        "STEP 0: stdscr={}:{}, win_row={}:{} initial cursor ={}:{}".format(
            term_row_max, term_col_max, win_row_max, win_col_max, y0, x0
        ),
        curses.color_pair(1),
    )
    stdscr.refresh()
    stdscr.getkey()
    #####################################################################################
    stdscr.clear()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
        curses.color_pair(1),
    )
    stdscr.addstr(term_row_max - 1, 0, "STEP 1: ", curses.color_pair(1))
    try:
        stdscr.addstr(term_row_max - 1, term_col_max - 1, "X", curses.color_pair(1))
    except curses.error as err:
        stdscr.addstr(
            term_row_max - 2,
            0,
            "err={}".format(err),
            curses.color_pair(1) | curses.A_BOLD,
        )
        stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    #####################################################################################
    popup = curses.newpad(300, 200)
    test_lines = test_text("@", 50, 200)
    for i, test_line in enumerate(test_lines.split("\n")):
        if i % 3 == 0:
            popup.addstr(i, 0, test_line[:win_col_max], curses.color_pair(2))
    popup.refresh(0, 0, margin, margin, win_row_max, win_col_max)
    stdscr.addstr(
        term_row_max - 2, 0, "NNNNNNNNNNNNNNNNNNNNNNNNNNN", curses.color_pair(2)
    )
    stdscr.addstr(term_row_max - 1, 0, "STEP 2: ", curses.color_pair(2))
    stdscr.refresh()
    stdscr.getkey()
    #####################################################################################
    margin_row = get_center_margin(term_row_max - 1, 12, 5)
    margin_col = get_center_margin(term_col_max, 60, 5)
    outbox = curses.newwin(8, 40, margin_row, margin_col)
    outbox_row_max, outbox_col_max = outbox.getmaxyx()
    outbox.border()
    outbox.refresh()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV",
        curses.color_pair(3),
    )
    stdscr.addstr(term_row_max - 1, 0, "STEP 3: ", curses.color_pair(3))
    stdscr.clrtoeol()
    stdscr.getkey()
    #####################################################################################
    inbox = curses.newwin(6, 36, margin_row + 1, margin_col + 2)
    # inbox = outbox.derwin(6, 36, 1, 2)
    inbox_row_max, inbox_col_max = inbox.getmaxyx()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "========================================================",
        curses.color_pair(4),
    )
    inbox.addstr(
        0,
        0,
        "Hello, world! ################################",
        curses.color_pair(4) | curses.A_REVERSE,
    )
    stdscr.clrtoeol()
    stdscr.addstr(term_row_max - 1, 0, "STEP 4: ", curses.color_pair(4))
    stdscr.clrtoeol()
    inbox.refresh()
    stdscr.getkey()
    #####################################################################################
    stdscr.addstr(
        term_row_max - 2,
        0,
        "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
        curses.color_pair(5),
    )
    stdscr.clrtoeol()
    stdscr.addstr(
        term_row_max - 1,
        0,
        "STEP 5: stdscr={}:{}, outbox={}:{}, inbox={}:{}".format(
            term_row_max,
            term_col_max,
            outbox_row_max,
            outbox_col_max,
            inbox_row_max,
            inbox_col_max,
        ),
        curses.color_pair(5),
    )
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "--------------------------------------------------------",
        curses.color_pair(6),
    )
    stdscr.addstr(term_row_max - 1, 0, "STEP 6: ", curses.color_pair(6))
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "++++++++++++++++++++++++++++++++++++++++++++++++++++++++",
        curses.color_pair(7),
    )
    stdscr.addstr(term_row_max - 1, 0, "STEP 7: ", curses.color_pair(7))
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(8) LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(8),
    )
    stdscr.clrtoeol()
    stdscr.addstr(term_row_max - 1, 0, "STEP 8: ", curses.color_pair(8))
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(1)|curses.A_REVERSE LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(1) | curses.A_REVERSE,
    )
    stdscr.clrtoeol()
    stdscr.addstr(
        term_row_max - 1, 0, "STEP 9: ", curses.color_pair(1) | curses.A_REVERSE
    )
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(1)|curses.A_BOLD LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(1) | curses.A_BOLD,
    )
    stdscr.clrtoeol()
    stdscr.addstr(
        term_row_max - 1, 0, "STEP 10: ", curses.color_pair(1) | curses.A_BOLD
    )
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(1)|curses.A_DIM LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(1) | curses.A_DIM,
    )
    stdscr.clrtoeol()
    stdscr.addstr(term_row_max - 1, 0, "STEP 11: ", curses.color_pair(1) | curses.A_DIM)
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 2,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(1)|curses.A_NORMAL LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(1) | curses.A_NORMAL,
    )
    stdscr.addstr(
        term_row_max - 1, 0, "STEP 12: ", curses.color_pair(1) | curses.A_NORMAL
    )
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    stdscr.addstr(
        term_row_max - 3,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(3)|curses.A_REVERSE LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(3) | curses.A_REVERSE,
    )
    stdscr.addstr(
        term_row_max - 2,
        0,
        "LLLLLLLLLLLLLLLLLLL curses.color_pair(3)|curses.A_DIM|curses.A_REVERSE LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL",
        curses.color_pair(3) | curses.A_DIM | curses.A_REVERSE,
    )
    stdscr.clrtoeol()
    stdscr.addstr(
        term_row_max - 1,
        0,
        "STEP 13: ",
        curses.color_pair(3) | curses.A_DIM | curses.A_REVERSE,
    )
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    long_line = "LLLLLLLLLLLLLLLLLLL curses.color_pair(4) | curses.A_REVERSE | curses.A_BOLD LLLLLLLLLLLLLLL LONG LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL"
    stdscr.move(term_row_max - 1, 0)
    y0, x0 = stdscr.getyx()  # don't forget () here
    try:
        stdscr.addstr(
            term_row_max - 1,
            0,
            long_line,
            curses.color_pair(4) | curses.A_REVERSE | curses.A_BOLD,
        )
    except curses.error as _:
        pass
    y1, x1 = stdscr.getyx()

    stdscr.addstr(
        term_row_max - 6,
        0,
        "STEP 14: before {}:{} len={}".format(y0, x0, len(long_line)),
        curses.color_pair(4) | curses.A_BOLD,
    )
    stdscr.addstr(
        term_row_max - 5,
        0,
        "STEP 14: after {}:{} max_col-1={}".format(y1, x1, term_col_max - 1),
        curses.color_pair(4) | curses.A_BOLD,
    )
    stdscr.refresh()
    stdscr.getkey()

    # stdscr.move(term_row_max - 1, term_col_max-1)
    stdscr.move(0, 0)
    y2, x2 = stdscr.getyx()
    try:
        stdscr.move(term_row_max - 1, term_col_max)
    except curses.error as _:
        pass
    y3, x3 = stdscr.getyx()
    try:
        stdscr.addstr(
            "@",
            curses.color_pair(6) | curses.A_REVERSE | curses.A_BOLD,
        )
    except curses.error as _:
        pass
    y4, x4 = stdscr.getyx()
    stdscr.clrtoeol()
    y5, x5 = stdscr.getyx()
    stdscr.addstr(
        term_row_max - 6,
        0,
        "STEP 15 at right-bottom after {}:{} >> outside {}:{} >>@ {}:{}  >>clrtoeol {}:{}  max={}:{}".format(
            y2, x2, y3, x3, y4, x4, y5, x5, term_row_max, term_col_max
        ),
        curses.color_pair(4) | curses.A_BOLD,
    )
    stdscr.clrtoeol()
    stdscr.refresh()
    stdscr.getkey()
    #####################################################################################


def get_center_margin(span_out, span_in, margin_min):
    span_in_real = min(span_out - 2 * margin_min, span_in)
    margin = (span_out - span_in_real) // 2
    return margin


####################################################################
# Internally used utility methods (popup display via curses)
####################################################################
def tui_popup_text(stdscr, text):
    curses.curs_set(0)
    term_row_max, term_col_max = stdscr.getmaxyx()
    stdscr.border()
    for line_index, line_text in enumerate(text.split("\n")):  # no \n
        if line_index < term_row_max - 5:
            stdscr.addstr(
                line_index + 1,
                2,
                line_text[: term_col_max - 5],
                curses.A_BOLD | curses.color_pair(7),
            )
    stdscr.refresh()
    stdscr.getch()
    return


if __name__ == "__main__":
    # run program
    main()
    # finish
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :


class Test:
    """TEST CODE"""

    def __init__(self):
        pass
        return

    def report(self, foo):  # To Be Overriden
        print(foo)
        return

    def main(self):
        self.report("HELLO")
        return


T = Test()

print("=====")
T.main()
print("=====")
T.report("OOPS")
print("=====")

# This allows different reporting style for TUI/CLI
